###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import annotations

import pytest

from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools.LogAnalysis.job_analyzer import (
    Job,
    PreAnalysisResult,
)

from . import load_example

TEST_CASES = {
    893788994: {
        "LFN:/lhcb/LHCb/Collision11/DIMUON.DST/00041838/0000/00041838_00003339_1.dimuon.dst": (
            PreAnalysisResult.NO_EVENTS_SELECTED
        ),
    },
    957694667: {
        "LFN:/lhcb/MC/2018/ALLSTREAMS.MDST/00097909/0000/00097909_00000169_7.AllStreams.mdst": (
            PreAnalysisResult.NEEDS_LOG_ANALYSIS
        )
    },
    893543510: {
        "LFN:/lhcb/LHCb/Collision12/LEPTONIC.MDST/00094006/0001/00094006_00018816_1.leptonic.mdst": (
            PreAnalysisResult.NO_XML_SUMMARY
        ),
        "LFN:/lhcb/LHCb/Collision12/LEPTONIC.MDST/00094006/0001/00094006_00019023_1.leptonic.mdst": (
            PreAnalysisResult.NO_XML_SUMMARY
        ),
        "LFN:/lhcb/LHCb/Collision12/LEPTONIC.MDST/00094006/0001/00094006_00019076_1.leptonic.mdst": (
            PreAnalysisResult.NO_XML_SUMMARY
        ),
        "LFN:/lhcb/LHCb/Collision12/LEPTONIC.MDST/00094006/0001/00094006_00019480_1.leptonic.mdst": (
            PreAnalysisResult.NO_XML_SUMMARY
        ),
        "LFN:/lhcb/LHCb/Collision12/LEPTONIC.MDST/00094006/0001/00094006_00019651_1.leptonic.mdst": (
            PreAnalysisResult.NO_XML_SUMMARY
        ),
        "LFN:/lhcb/LHCb/Collision12/LEPTONIC.MDST/00094006/0002/00094006_00020251_1.leptonic.mdst": (
            PreAnalysisResult.NO_XML_SUMMARY
        ),
        "LFN:/lhcb/LHCb/Collision12/LEPTONIC.MDST/00094006/0002/00094006_00020785_1.leptonic.mdst": (
            PreAnalysisResult.NO_XML_SUMMARY
        ),
        "LFN:/lhcb/LHCb/Collision12/LEPTONIC.MDST/00094006/0002/00094006_00021863_1.leptonic.mdst": (
            PreAnalysisResult.NO_XML_SUMMARY
        ),
        "LFN:/lhcb/LHCb/Collision12/LEPTONIC.MDST/00094006/0002/00094006_00024696_1.leptonic.mdst": (
            PreAnalysisResult.NO_XML_SUMMARY
        ),
        "LFN:/lhcb/LHCb/Collision12/LEPTONIC.MDST/00094006/0002/00094006_00024886_1.leptonic.mdst": (
            PreAnalysisResult.NO_XML_SUMMARY
        ),
    },
    957012280: {
        "LFN:/lhcb/MC/2016/ALLRADIATIVE.STRIP.DST/00154282/0000/00154282_00000001_1.allradiative.strip.dst": (
            PreAnalysisResult.MAYBE_POST_APPLICATION_FAILURE
        ),
    },
    893803907: {
        "LFN:/lhcb/LHCb/Collision24/CHARM.DST/00226358/0012/00226358_00121543_1.charm.dst": PreAnalysisResult.NO_XML_SUMMARY,
        "LFN:/lhcb/LHCb/Collision24/CHARM.DST/00226358/0012/00226358_00121830_1.charm.dst": PreAnalysisResult.NO_XML_SUMMARY,
        "LFN:/lhcb/LHCb/Collision24/CHARM.DST/00226358/0012/00226358_00121831_1.charm.dst": PreAnalysisResult.NO_XML_SUMMARY,
        "LFN:/lhcb/LHCb/Collision24/CHARM.DST/00226358/0012/00226358_00121838_1.charm.dst": PreAnalysisResult.NO_XML_SUMMARY,
        "LFN:/lhcb/LHCb/Collision24/CHARM.DST/00226358/0012/00226358_00121839_1.charm.dst": PreAnalysisResult.NO_XML_SUMMARY,
        "LFN:/lhcb/LHCb/Collision24/CHARM.DST/00226358/0012/00226358_00122140_1.charm.dst": PreAnalysisResult.NO_XML_SUMMARY,
        "LFN:/lhcb/LHCb/Collision24/CHARM.DST/00226358/0012/00226358_00122141_1.charm.dst": PreAnalysisResult.NO_XML_SUMMARY,
        "LFN:/lhcb/LHCb/Collision24/CHARM.DST/00226358/0012/00226358_00122151_1.charm.dst": PreAnalysisResult.NO_XML_SUMMARY,
        "LFN:/lhcb/LHCb/Collision24/CHARM.DST/00226358/0012/00226358_00122471_1.charm.dst": PreAnalysisResult.NO_XML_SUMMARY,
        "LFN:/lhcb/LHCb/Collision24/CHARM.DST/00226358/0012/00226358_00122484_1.charm.dst": PreAnalysisResult.NO_XML_SUMMARY,
    },
    940635849: {
        "LFN:/lhcb/data/2023/RAW/BEAMGAS/LHCb/VDM23/275640/275640_00090015_0000.raw": (
            PreAnalysisResult.UNKNOWN_ERROR_ON_JOB_SUCCESS
        ),
        "LFN:/lhcb/data/2023/RAW/BEAMGAS/LHCb/VDM23/275640/275640_00090009_0000.raw": (
            PreAnalysisResult.UNKNOWN_ERROR_ON_JOB_SUCCESS
        ),
    },
    886255904: {
        "LFN:/lhcb/MC/2018/ALLSTREAMS.DST/00173490/0000/00173490_00001002_7.AllStreams.dst": PreAnalysisResult.NO_EVENTS_SELECTED,
        "LFN:/lhcb/MC/2018/ALLSTREAMS.DST/00173490/0000/00173490_00001009_7.AllStreams.dst": PreAnalysisResult.NO_EVENTS_SELECTED,
    },
    929522350: {
        "LFN:/lhcb/LHCb/Collision24/B2CC.DST/00231024/0009/00231024_00092637_1.b2cc.dst": (
            PreAnalysisResult.MAYBE_POST_APPLICATION_FAILURE
        ),
        "LFN:/lhcb/LHCb/Collision24/B2CC.DST/00231024/0009/00231024_00098545_1.b2cc.dst": (
            PreAnalysisResult.MAYBE_POST_APPLICATION_FAILURE
        ),
        "LFN:/lhcb/LHCb/Collision24/B2CC.DST/00231024/0009/00231024_00098828_1.b2cc.dst": (
            PreAnalysisResult.MAYBE_POST_APPLICATION_FAILURE
        ),
        "LFN:/lhcb/LHCb/Collision24/B2CC.DST/00231024/0009/00231024_00098974_1.b2cc.dst": (
            PreAnalysisResult.MAYBE_POST_APPLICATION_FAILURE
        ),
        "LFN:/lhcb/LHCb/Collision24/B2CC.DST/00231024/0009/00231024_00099072_1.b2cc.dst": (
            PreAnalysisResult.MAYBE_POST_APPLICATION_FAILURE
        ),
        "LFN:/lhcb/LHCb/Collision24/B2CC.DST/00231024/0009/00231024_00099474_1.b2cc.dst": (
            PreAnalysisResult.MAYBE_POST_APPLICATION_FAILURE
        ),
        "LFN:/lhcb/LHCb/Collision24/B2CC.DST/00231024/0010/00231024_00104352_1.b2cc.dst": (
            PreAnalysisResult.MAYBE_POST_APPLICATION_FAILURE
        ),
        "LFN:/lhcb/LHCb/Collision24/B2CC.DST/00231024/0010/00231024_00104414_1.b2cc.dst": (
            PreAnalysisResult.MAYBE_POST_APPLICATION_FAILURE
        ),
        "LFN:/lhcb/LHCb/Collision24/B2CC.DST/00231024/0010/00231024_00105096_1.b2cc.dst": (
            PreAnalysisResult.MAYBE_POST_APPLICATION_FAILURE
        ),
        "LFN:/lhcb/LHCb/Collision24/B2CC.DST/00231024/0015/00231024_00156493_1.b2cc.dst": (
            PreAnalysisResult.MAYBE_POST_APPLICATION_FAILURE
        ),
    },
    895008813: {
        "LFN:/lhcb/LHCb/Collision24/BANDQ.DST/00227656/0007/00227656_00075025_1.bandq.dst": PreAnalysisResult.PROBABLY_OK,
        "LFN:/lhcb/LHCb/Collision24/BANDQ.DST/00227656/0007/00227656_00074480_1.bandq.dst": PreAnalysisResult.PROBABLY_OK,
        "LFN:/lhcb/LHCb/Collision24/BANDQ.DST/00227656/0006/00227656_00067693_1.bandq.dst": PreAnalysisResult.PROBABLY_OK,
        "LFN:/lhcb/LHCb/Collision24/BANDQ.DST/00227656/0006/00227656_00067922_1.bandq.dst": PreAnalysisResult.PROBABLY_OK,
        "LFN:/lhcb/LHCb/Collision24/BANDQ.DST/00227656/0006/00227656_00068205_1.bandq.dst": PreAnalysisResult.PROBABLY_OK,
        "LFN:/lhcb/LHCb/Collision24/BANDQ.DST/00227656/0006/00227656_00068589_1.bandq.dst": PreAnalysisResult.PROBABLY_OK,
        "LFN:/lhcb/LHCb/Collision24/BANDQ.DST/00227656/0006/00227656_00068624_1.bandq.dst": PreAnalysisResult.NOT_IN_SUMMARY_XML,
        "LFN:/lhcb/LHCb/Collision24/BANDQ.DST/00227656/0006/00227656_00068625_1.bandq.dst": PreAnalysisResult.NOT_IN_SUMMARY_XML,
        "LFN:/lhcb/LHCb/Collision24/BANDQ.DST/00227656/0006/00227656_00068843_1.bandq.dst": PreAnalysisResult.NOT_IN_SUMMARY_XML,
        "LFN:/lhcb/LHCb/Collision24/BANDQ.DST/00227656/0006/00227656_00068623_1.bandq.dst": PreAnalysisResult.NEEDS_LOG_ANALYSIS,
    },
    # TODO: Add a case of PreAnalysisResult.NEEDS_GROUP_SIZE_REDUCTION
}


@pytest.mark.parametrize("job_id, expected_result", TEST_CASES.items())
def test_pre_analysis(job_id: int, expected_result: dict[str, set[str]]):
    transform_id, task_id, files = load_example(job_id)
    job = Job(transform_id, task_id, files)
    assert job.pre_analysis == expected_result



###############################################################################
# (c) Copyright 2025 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import annotations

__all__ = ["analyse_log"]

import hyperscan

from .reasons import GenericPythonCrash  # pylint: disable=no-name-in-module
from .job_analyzer import ReasonInfo


db = hyperscan.Database()
patterns = ((GenericPythonCrash.pattern, 0, GenericPythonCrash.pattern_flags),)
expressions, ids, flags = zip(*patterns)
db.compile(expressions=expressions, ids=ids, elements=len(patterns), flags=flags)


def analyse_log(log) -> list[ReasonInfo]:
    context = []
    db.scan(log, match_event_handler=_on_match, context=(log, context))
    return context


def _on_match(id: int, start: int, end: int, flags: int, context: tuple[bytes, list[ReasonInfo]]) -> bool | None:
    reason_info = ReasonInfo(
        type="GenericPythonCrash",
        match=context[0][start:end].decode(errors="backslashreplace"),
        explanation=GenericPythonCrash.explanation,
    )
    context[1].append(reason_info)



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import annotations

import json
import os

import pytest
from pydantic import TypeAdapter

from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools.LogAnalysis.job_analyzer import Job, LFNDebugInfo

from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools.LogAnalysis.test import (
    deterministic_dump,
    generate_fixture_jobid,
    load_example,
    EXAMPLE_LOGS_DIR,
)

TEST_CASES = [int(x.name.removesuffix(".tar.zst")) for x in EXAMPLE_LOGS_DIR.glob("*.tar.zst")]
PreAnalysisTypeAdapter = TypeAdapter(dict[str, LFNDebugInfo])

UPDATE_REFERENCES_ENV_VAR = "LHCBDIRAC_TESTING_LOG_ANALYSIS_UPDATE_REFERENCES"
TEST_MODULE_NAME = __name__.rsplit(".", 1)[-1]  # "Test_log_analysis"


@pytest.mark.parametrize("job_id", TEST_CASES)
def test_log_analysis(job_id: int):
    expected_result_path = EXAMPLE_LOGS_DIR / f"{job_id}.json"

    transform_id, task_id, files = load_example(job_id)
    job = Job(transform_id, task_id, files)
    actual_result = PreAnalysisTypeAdapter.dump_python(job.problems_by_lfn)

    if os.environ.get(UPDATE_REFERENCES_ENV_VAR):
        expected_result_path.write_text(deterministic_dump(PreAnalysisTypeAdapter, job.problems_by_lfn))
        pytest.skip(f"Updated reference file {expected_result_path}")

    if not expected_result_path.is_file():
        pytest.fail(
            f"Reference file {expected_result_path} does not exist.\n"
            f"Set {UPDATE_REFERENCES_ENV_VAR}=1 to create it."
        )

    expected_result = json.loads(expected_result_path.read_text())

    if actual_result != expected_result:
        pytest.fail(
            f"Result mismatch for job {job_id}.\n" f"Set {UPDATE_REFERENCES_ENV_VAR}=1 to update the reference file."
        )


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--jobid", type=int, help="Only generate reference results for this job ID")
    args = parser.parse_args()

    generate_fixture_jobid(args.jobid)



###############################################################################
# (c) Copyright 2025 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import requests


class CERNSSOAPIToken:
    def __init__(
        self,
        audience: str,
        *,
        client_id: str | None = None,
        client_secret: str | None = None,
    ):
        self.client_id = client_id
        self.client_secret = client_secret
        self.audience = audience

    @property
    def token(self):
        """Exchange for an api token for the given audience.
        See: https://auth.docs.cern.ch/user-documentation/oidc/api-access/
        """
        response = requests.post(
            "https://auth.cern.ch/auth/realms/cern/api-access/token",
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            data={
                "grant_type": "client_credentials",
                "client_id": self.client_id,
                "client_secret": self.client_secret,
                "audience": self.audience,
            },
        )

        if response.status_code != 200:
            raise ValueError(f"Could not get API token from {self.audience=}: {response.text}")

        return response.json().get("access_token")



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Tests for the Pydantic-based reason configuration system."""
from __future__ import annotations

import re
import tempfile
from pathlib import Path

import pytest
from pydantic import ValidationError

from ..base_reasons import (
    WorkerNodeIssue,
    RecoverableFailure,
)
from ..reason_models import ReasonDefinition, ReasonsConfig


class TestReasonDefinition:
    """Tests for ReasonDefinition Pydantic model."""

    def test_valid_reason_definition(self):
        """Test that valid reason definitions work."""
        reason = ReasonDefinition(
            name="TestFailure",
            base="WorkerNodeIssue",
            patterns=["test pattern"],
            explanation="Test explanation",
        )
        assert reason.name == "TestFailure"
        assert reason.base == "WorkerNodeIssue"
        assert reason.patterns == ["test pattern"]
        assert reason.explanation == "Test explanation"

    def test_valid_reason_with_issue_url(self):
        """Test reason with issue URL."""
        reason = ReasonDefinition(
            name="TestFailure",
            base="KnownUnrecoverableFailure",
            patterns=["error pattern"],
            issue_url="https://gitlab.cern.ch/lhcb/test/-/issues/123",
        )
        assert reason.issue_url == "https://gitlab.cern.ch/lhcb/test/-/issues/123"

    def test_invalid_regex_pattern(self):
        """Test that invalid regex patterns are caught."""
        with pytest.raises(ValidationError, match="Invalid regex pattern"):
            ReasonDefinition(
                name="BadPattern",
                base="WorkerNodeIssue",
                patterns=["[invalid(regex"],  # Unclosed bracket
                explanation="Test",
            )

    def test_invalid_name_lowercase(self):
        """Test that names starting with lowercase are rejected."""
        with pytest.raises(ValidationError, match="must start with uppercase"):
            ReasonDefinition(
                name="badName",
                base="WorkerNodeIssue",
                patterns=["test"],
            )

    def test_multiple_patterns(self):
        """Test reason with multiple patterns."""
        reason = ReasonDefinition(
            name="MultiPattern",
            base="RecoverableFailure",
            patterns=["pattern1", "pattern2", "pattern3"],
            explanation="Multiple patterns test",
        )
        assert len(reason.patterns) == 3

    def test_to_class_single_pattern(self):
        """Test converting definition to class with single pattern."""
        reason = ReasonDefinition(
            name="SinglePattern",
            base="WorkerNodeIssue",
            patterns=["test pattern"],
            explanation="Test explanation",
        )
        cls = reason.to_class()

        assert cls.__name__ == "SinglePattern"
        assert issubclass(cls, WorkerNodeIssue)
        assert cls.pattern == b"test pattern"
        assert cls.explanation == "Test explanation"

    def test_to_class_multiple_patterns(self):
        """Test converting definition to class with multiple patterns."""
        reason = ReasonDefinition(
            name="MultiPattern",
            base="RecoverableFailure",
            patterns=["pattern1", "pattern2"],
            explanation="Test",
        )
        cls = reason.to_class()

        assert cls.__name__ == "MultiPattern"
        assert issubclass(cls, RecoverableFailure)
        assert cls.patterns == [b"pattern1", b"pattern2"]

    def test_mixin_specification(self):
        """Test mixin specification."""
        reason = ReasonDefinition(
            name="CustomReason",
            base="UnknownFailure",
            patterns=["test"],
            mixin=".reasons:FatalGaudiErrorMixin",
        )
        assert reason.mixin == ".reasons:FatalGaudiErrorMixin"


class TestReasonsConfig:
    """Tests for ReasonsConfig Pydantic model."""

    def test_valid_config(self):
        """Test valid configuration."""
        config = ReasonsConfig(
            reasons=[
                ReasonDefinition(
                    name="Reason1",
                    base="WorkerNodeIssue",
                    patterns=["pattern1"],
                ),
                ReasonDefinition(
                    name="Reason2",
                    base="RecoverableFailure",
                    patterns=["pattern2"],
                ),
            ]
        )
        assert len(config.reasons) == 2

    def test_duplicate_reason_names(self):
        """Test that duplicate names are rejected."""
        with pytest.raises(ValidationError, match="Duplicate reason names"):
            ReasonsConfig(
                reasons=[
                    ReasonDefinition(name="Duplicate", base="WorkerNodeIssue", patterns=["a"]),
                    ReasonDefinition(name="Duplicate", base="RecoverableFailure", patterns=["b"]),
                ]
            )

    def test_load_from_yaml(self):
        """Test loading configuration from YAML file."""
        yaml_content = """
reasons:
  - name: TestReason
    base: WorkerNodeIssue
    patterns: ["test pattern"]
    explanation: "Test explanation"
  - name: AnotherReason
    base: KnownUnrecoverableFailure
    patterns: ["another pattern"]
    issue_url: "https://gitlab.cern.ch/lhcb/test/-/issues/1"
"""
        with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
            f.write(yaml_content)
            yaml_file = Path(f.name)

        try:
            config = ReasonsConfig.load_from_yaml(yaml_file)
            assert len(config.reasons) == 2
            assert config.reasons[0].name == "TestReason"
            assert config.reasons[1].name == "AnotherReason"
            assert config.reasons[1].issue_url == "https://gitlab.cern.ch/lhcb/test/-/issues/1"
        finally:
            yaml_file.unlink()

    def test_create_reason_classes(self):
        """Test that classes are created correctly."""
        config = ReasonsConfig(
            reasons=[
                ReasonDefinition(
                    name="TestFailure",
                    base="WorkerNodeIssue",
                    patterns=["test pattern"],
                    explanation="Test explanation",
                ),
                ReasonDefinition(
                    name="CustomFailure",
                    base="UnknownFailure",
                    patterns=["custom"],
                    mixin=".reasons:FatalGaudiErrorMixin",
                ),
            ]
        )

        classes = config.create_reason_classes()

        # Both classes should be created (mixins are composed in class hierarchy)
        assert "TestFailure" in classes
        assert "CustomFailure" in classes

        TestFailure = classes["TestFailure"]
        assert TestFailure.pattern == b"test pattern"
        assert TestFailure.explanation == "Test explanation"
        assert issubclass(TestFailure, WorkerNodeIssue)

        # Check that CustomFailure has the mixin in its MRO
        CustomFailure = classes["CustomFailure"]
        from ..reasons import FatalGaudiErrorMixin

        assert FatalGaudiErrorMixin in CustomFailure.__mro__

    def test_empty_reasons_list(self):
        """Test configuration with empty reasons list."""
        # This should fail validation since we require at least one reason
        # But let's test an empty list to ensure it doesn't crash
        config = ReasonsConfig(reasons=[])
        assert len(config.reasons) == 0
        classes = config.create_reason_classes()
        assert len(classes) == 0


class TestYAMLIntegration:
    """Integration tests for YAML loading and class creation."""

    def test_load_actual_reasons_yaml(self):
        """Test loading the actual reasons.yaml file."""
        # This tests that the real configuration file is valid
        config = ReasonsConfig.load_from_yaml()
        assert len(config.reasons) > 0

        # Check that we have a mix of classes with and without mixins
        mixin_count = sum(1 for r in config.reasons if r.mixin)
        regular_count = len(config.reasons) - mixin_count

        assert regular_count > 0, "Should have regular YAML-defined reasons"
        assert mixin_count >= 0, "May have reasons with mixins"

    def test_create_classes_from_actual_yaml(self):
        """Test creating classes from the actual YAML file."""
        config = ReasonsConfig.load_from_yaml()
        classes = config.create_reason_classes()

        # Should have created multiple classes
        assert len(classes) > 0

        # Verify some expected classes exist
        assert "FunctorCompilationFailure" in classes or any(
            "Functor" in name for name in classes
        ), "Should have at least one reason class"

        # Pick the first class and verify it has the expected attributes
        first_class_name = next(iter(classes.keys()))
        first_class = classes[first_class_name]

        assert hasattr(first_class, "pattern") or hasattr(first_class, "patterns")
        assert hasattr(first_class, "explanation") or first_class.explanation is None

    def test_patterns_are_valid_regex(self):
        """Test that all patterns in the YAML are valid regex."""
        config = ReasonsConfig.load_from_yaml()

        for reason in config.reasons:
            for pattern in reason.patterns:
                # This should not raise an exception
                try:
                    print(f"Testing regex pattern: {pattern!r} for reason: {reason.name}")
                    re.compile(pattern.encode())
                except re.error as e:
                    pytest.fail(f"Invalid regex in {reason.name}: {pattern} - {e}")

    def test_base_classes_are_valid(self):
        """Test that all base classes are valid."""
        config = ReasonsConfig.load_from_yaml()

        valid_bases = {
            "WorkerNodeIssue",
            "RecoverableFailure",
            "KnownUnrecoverableFailure",
            "EventTimeout",
            "RecoverableFailureWithManualIntervention",
            "UnknownFailure",
        }

        for reason in config.reasons:
            assert reason.base in valid_bases, f"Invalid base class: {reason.base}"


class TestReasonModule:
    """Tests for the reasons module that loads from YAML."""

    def test_import_yaml_defined_classes(self):
        """Test that YAML-defined classes can be imported."""
        # Import the module
        from .. import reasons

        # Check that we have the expected classes
        assert hasattr(reasons, "FatalGaudiError")
        assert hasattr(reasons, "UnknownSegmentationFaultWithTraceback")
        assert hasattr(reasons, "KnownCorruptedFile")

        # Check that we have at least some YAML-defined classes
        # We know these exist in the YAML
        expected_yaml_classes = [
            "FunctorCompilationFailure",
            "FileTooLarge",
            "EventWatchdog",
        ]

        for class_name in expected_yaml_classes:
            if hasattr(reasons, class_name):
                # Found at least one
                break
        else:
            pytest.fail(f"None of the expected YAML classes found: {expected_yaml_classes}")

    def test_custom_classes_have_mixins(self):
        """Test that classes with mixins have their custom methods."""
        from .. import reasons

        # FatalGaudiError should have the FatalGaudiErrorMixin in its MRO
        assert hasattr(reasons.FatalGaudiError, "match_string")  # pylint: disable=no-member
        assert reasons.FatalGaudiErrorMixin in reasons.FatalGaudiError.__mro__  # pylint: disable=no-member

        # UnknownSegmentationFaultWithTraceback should have SegfaultTracebackMixin
        assert hasattr(reasons.UnknownSegmentationFaultWithTraceback, "match_string")  # pylint: disable=no-member
        assert (
            reasons.SegfaultTracebackMixin
            in reasons.UnknownSegmentationFaultWithTraceback.__mro__  # pylint: disable=no-member
        )

    def test_all_exports(self):
        """Test that __all__ is properly defined."""
        from .. import reasons

        assert hasattr(reasons, "__all__")
        assert isinstance(reasons.__all__, list)
        assert len(reasons.__all__) > 0

        # Check that custom classes are in __all__
        assert "FatalGaudiError" in reasons.__all__
        assert "UnknownSegmentationFaultWithTraceback" in reasons.__all__
        assert "KnownCorruptedFile" in reasons.__all__



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Pydantic models for defining failure reasons via YAML configuration."""
from __future__ import annotations

import os
import re
from pathlib import Path
from typing import Literal

import yaml
from pydantic import BaseModel, Field, field_validator, model_validator

from .base_reasons import (
    WorkerNodeIssue,
    RecoverableFailure,
    KnownUnrecoverableFailure,
    EventTimeout,
    RecoverableFailureWithManualIntervention,
    UnknownFailure,
    AppLogFailureReason,
)


# Type mapping for base classes
BaseReasonType = Literal[
    "WorkerNodeIssue",
    "RecoverableFailure",
    "KnownUnrecoverableFailure",
    "EventTimeout",
    "RecoverableFailureWithManualIntervention",
    "UnknownFailure",
]

BASE_REASON_REGISTRY = {
    "WorkerNodeIssue": WorkerNodeIssue,
    "RecoverableFailure": RecoverableFailure,
    "KnownUnrecoverableFailure": KnownUnrecoverableFailure,
    "EventTimeout": EventTimeout,
    "RecoverableFailureWithManualIntervention": RecoverableFailureWithManualIntervention,
    "UnknownFailure": UnknownFailure,
}


class ReasonDefinition(BaseModel):
    """Pydantic model for a single reason definition from YAML."""

    name: str = Field(..., description="Class name for this reason")
    base: BaseReasonType = Field(..., description="Base class to inherit from")
    patterns: list[str] = Field(..., min_length=1, description="List of regex patterns (as strings)")
    explanation: str | None = Field(None, description="Human-readable explanation of the failure")
    issue_url: str | None = Field(None, description="URL to GitLab issue or documentation")
    mixin: str | None = Field(
        None,
        description="Mixin class to customize behavior (format: 'module.path:ClassName' or '.relative.module:ClassName')",
    )
    pattern_flags: int | None = Field(None, description="Optional hyperscan pattern flags")

    @field_validator("patterns")
    @classmethod
    def validate_patterns(cls, v: list[str]) -> list[str]:
        """Validate that patterns are valid regex and compile as bytes."""
        for i, pattern in enumerate(v, 1):
            try:
                # Compile as bytes since that's how they're used in the application
                re.compile(pattern.encode())
            except re.error as e:
                # Provide detailed error message with pattern preview
                pattern_preview = pattern[:100] + "..." if len(pattern) > 100 else pattern
                raise ValueError(f"Invalid regex pattern #{i}: {pattern_preview!r}\n" f"Regex error: {e}") from e
        return v

    @field_validator("name")
    @classmethod
    def validate_name(cls, v: str) -> str:
        """Ensure name follows Python class naming conventions."""
        if not v[0].isupper():
            raise ValueError(f"Reason name must start with uppercase: {v}")
        if not v.replace("_", "").replace(".", "").isalnum():
            raise ValueError(f"Reason name must be alphanumeric: {v}")
        return v

    @model_validator(mode="after")
    def validate_issue_url_for_known_failures(self) -> ReasonDefinition:
        """Warn if KnownUnrecoverableFailure lacks an issue_url."""
        if self.base == "KnownUnrecoverableFailure" and not self.issue_url and not self.mixin:
            # This is a soft validation - just for awareness
            pass
        return self

    def to_class(self) -> type[AppLogFailureReason]:
        """Convert this definition to an actual Python class.

        If a mixin is specified, it will be loaded and included in the class hierarchy.
        The mixin is placed before the base class in the MRO to allow method overriding.
        """
        base_class = BASE_REASON_REGISTRY[self.base]

        # Build class attributes
        class_attrs = {}

        # Add patterns (convert strings to bytes)
        patterns_bytes = [p.encode() for p in self.patterns]
        if len(patterns_bytes) == 1:
            class_attrs["pattern"] = patterns_bytes[0]
        else:
            class_attrs["patterns"] = patterns_bytes

        # Add optional attributes
        if self.explanation is not None:
            class_attrs["explanation"] = self.explanation
        if self.issue_url is not None:
            class_attrs["issue_url"] = self.issue_url
        if self.pattern_flags is not None:
            class_attrs["pattern_flags"] = self.pattern_flags

        # Load mixin if specified
        if self.mixin:
            mixin_class = self._load_mixin(self.mixin)
            # Create class with mixin before base class in MRO
            return type(self.name, (mixin_class, base_class), class_attrs)
        else:
            # Create class without mixin
            return type(self.name, (base_class,), class_attrs)

    def _load_mixin(self, mixin_path: str) -> type:
        """Load a mixin class from a module path.

        Args:
            mixin_path: String in format "module.path:ClassName" or ".relative.module:ClassName"

        Returns:
            The mixin class

        Raises:
            ValueError: If the path format is invalid
            ImportError: If the module cannot be imported
            AttributeError: If the class doesn't exist in the module
        """
        if ":" not in mixin_path:
            raise ValueError(f"Invalid mixin path format: {mixin_path}. Expected 'module:ClassName'")

        module_path, class_name = mixin_path.rsplit(":", 1)

        # Import the module
        if module_path.startswith("."):
            from importlib import import_module

            module = import_module(
                module_path, package="LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools.LogAnalysis"
            )
        else:
            import importlib

            module = importlib.import_module(module_path)

        # Get the mixin class
        mixin_class = getattr(module, class_name)

        # Validate that it's a class
        if not isinstance(mixin_class, type):
            raise TypeError(f"{mixin_path} is not a class")

        return mixin_class


class ReasonsConfig(BaseModel):
    """Root configuration model for all reasons."""

    reasons: list[ReasonDefinition] = Field(..., description="List of all reason definitions")

    @field_validator("reasons")
    @classmethod
    def validate_unique_names(cls, v: list[ReasonDefinition]) -> list[ReasonDefinition]:
        """Ensure all reason names are unique."""
        names = [r.name for r in v]
        if len(names) != len(set(names)):
            duplicates = [name for name in names if names.count(name) > 1]
            raise ValueError(f"Duplicate reason names found: {duplicates}")
        return v

    @classmethod
    def load_from_yaml(cls, path: Path | None = None) -> ReasonsConfig:
        """Load configuration from YAML file."""
        if path is None:
            path = Path(__file__).parent / "reasons.yaml"
            if "DIRAC_PROD_MANAGER_REASONS_PATH" in os.environ:
                path = Path(os.environ["DIRAC_PROD_MANAGER_REASONS_PATH"])

        with open(path) as f:
            data = yaml.safe_load(f)

        return cls.model_validate(data)

    def create_reason_classes(self) -> dict[str, type[AppLogFailureReason]]:
        """Create reason classes from this configuration.

        Returns:
            Dictionary mapping reason names to their class objects
        """
        return {reason_def.name: reason_def.to_class() for reason_def in self.reasons}



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import annotations

import fnmatch
import json
import tarfile
from pathlib import Path

import zstandard as zstd


EXAMPLE_LOGS_DIR = Path(__file__).parent / "examples" / "logs"


def load_example(job_id: int) -> tuple[int, int, dict[str, bytes]]:
    """Load an example log file for the given job ID."""
    archive_path = EXAMPLE_LOGS_DIR / f"{job_id}.tar.zst"

    # If compressed archive exists, extract files from it
    if archive_path.exists():
        files = {}
        # Decompress and read files directly from the archive without writing to disk
        with open(archive_path, "rb") as compressed:
            dctx = zstd.ZstdDecompressor()
            with dctx.stream_reader(compressed) as reader:
                with tarfile.open(fileobj=reader, mode="r|") as tar:
                    for member in tar:
                        if member.isfile():
                            # Extract just the filename from the path (removes job_id directory prefix)
                            filename = Path(member.name).name
                            file_obj = tar.extractfile(member)
                            if file_obj:
                                files[filename] = file_obj.read()
    else:
        # Fall back to reading from uncompressed directory
        files = {}
        for file in (EXAMPLE_LOGS_DIR / f"{job_id}").iterdir():
            files[file.name] = file.read_bytes()

    # Extract the transform and task IDs from the prodConf filename
    *_, transform_id, task_id, _ = fnmatch.filter(files, "prodConf*_1.*")[0].split("_")
    return int(transform_id), int(task_id), files


def deterministic_dump(adapter, obj) -> str:
    # Workaround for https://github.com/pydantic/pydantic/issues/7424
    return json.dumps(adapter.dump_python(obj, mode="json"), indent=2, sort_keys=True)


def generate_fixture_jobid(jobid: int | None = None):
    from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools.LogAnalysis.job_analyzer import Job
    from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools.LogAnalysis.test.Test_log_analysis import (
        PreAnalysisTypeAdapter,
    )

    if jobid is not None:
        print(f"Generating fixture for job ID {jobid}")
        jobid = int(jobid)

    # Generate reference results from examples
    for path in EXAMPLE_LOGS_DIR.glob("*.tar.zst"):
        job_id = int(path.name.removesuffix(".tar.zst"))
        if jobid is not None and job_id != jobid:
            continue
        transform_id, task_id, files = load_example(job_id)
        expected_path = EXAMPLE_LOGS_DIR / f"{job_id}.json"
        expected_path.write_text(
            deterministic_dump(PreAnalysisTypeAdapter, Job(transform_id, task_id, files).problems_by_lfn)
        )
        print(f"Wrote {expected_path}")



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbDIRAC.ResourceStatusSystem.Agent.ShiftDBAgent.

ShiftDBAgent.__bases__:
  DIRAC.Core.Base.AgentModule.AgentModule
"""
# FIXME: should add a "DryRun" option to run in certification setup
from urllib.request import urlopen
from urllib.error import URLError
import json
import requests

from DIRAC import gConfig, S_OK, S_ERROR
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

AGENT_NAME = "ResourceStatus/ShiftDBAgent"


class ShiftDBAgent(AgentModule):
    """This agent queries the LHCb ShiftDB and gets the emails of each piquet
    Then, populates the eGroup associated
    The e-groups admin should be : lhcb-grid-experiment-egroup-admins
    """

    def __init__(self, *args, **kwargs):
        AgentModule.__init__(self, *args, **kwargs)

        # Members initialization

        # ShiftDB url where to find shifter emails
        self.lbshiftdburl = "https://lbshiftdb.cern.ch/list_email"

        self.roles = {}
        self.roleShifters = {}
        self.newShifters = {}

        self.diracAdmin = None

    def initialize(self, *args, **kwargs):
        """Initialize."""

        # === Configuration ===
        self.client_id = self.am_getOption("client_id")
        self.client_secret = self.am_getOption("client_secret")

        self.keycloak_api_token_endpoint = "https://auth.cern.ch/auth/realms/cern/api-access/token"
        self.authzsvc_endpoint = "https://authorization-service-api.web.cern.ch/api/v1.0/"

        self.lbshiftdburl = self.am_getOption("lbshiftdburl", self.lbshiftdburl)

        self.diracAdmin = DiracAdmin()

        return S_OK()

    def beginExecution(self):
        self.log.info("Getting roles from CS")
        _section = self.am_getModuleParam("section")
        self.roles = gConfig.getSections(f"{_section}/roles")
        self.eGroups = {}

        for role in self.roles["Value"]:
            eGroup = gConfig.getValue(f"{_section}/roles/{role}/eGroup")
            if eGroup:
                self.eGroups[role] = eGroup
                self.log.debug(f"Found {role} : {eGroup} ")

        return S_OK()

    def execute(self):
        """Execution."""

        #
        # === Get API Token ===
        self.log.info("Getting API token...")
        try:
            token_resp = requests.post(
                self.keycloak_api_token_endpoint,
                data={
                    "grant_type": "client_credentials",
                    "client_id": self.client_id,
                    "client_secret": self.client_secret,
                    "audience": "authorization-service-api",
                },
                headers={"Content-Type": "application/x-www-form-urlencoded"},
            )
            token_resp.raise_for_status()
            self.api_token = token_resp.json()["access_token"]
            self.log.info("Token received.\n")
        except Exception as e:
            self.log.error("Failed to get API token:", e)

        self.roleShifters = {}
        self.newShifters = {}

        self.log.info("Getting role emails")

        for role, eGroup in self.eGroups.items():
            self.roles[role] = eGroup

            ccid = self.__getRoleCCID(role)
            if not ccid["OK"]:
                self.log.error(ccid["Message"])
                # We do not return, we keep execution to clean old shifters
                ccid["Value"] = None

            ccid = ccid["Value"]
            self.roleShifters[eGroup] = (ccid, role)

            self.log.info(f"{role} -> {ccid}")

        self.log.info("Setting role CCIDs")
        for eGroup, roleTuple in self.roleShifters.items():
            ccid, role = roleTuple
            if ccid is not None:
                setCCID = self.__setRoleCCID(eGroup, ccid, role)
                if not setCCID["OK"]:
                    self.log.error(setCCID["Message"])

        for newShifterRole, shifterEgroup in self.newShifters.items():
            self.log.info(f"Notifying role {newShifterRole}")
            res = self.__notifyNewShifter(newShifterRole, shifterEgroup)
            if not res["OK"]:
                self.log.error(res["Message"])

        return S_OK()

    def __getRoleCCID(self, role):
        """Get role CCID from shiftDB."""

        try:
            web = urlopen(self.lbshiftdburl, timeout=60)
        except URLError as e:
            return S_ERROR(f"Cannot open URL: {self.lbshiftdburl}, error {e}")

        CCIDperson = []
        listCCID = []

        for line in web.readlines():
            for item in json.loads(line):
                if role in item["role"]:
                    # There are three shifts per day, so we take into account what time is it
                    # before sending the email.
                    CCIDperson = {"id": (self.__getEmailCCID(item["email"])["data"][0]["personId"])}
                    listCCID.append(CCIDperson)

        if not listCCID:
            return S_ERROR("CCID not found")

        return S_OK(listCCID)

    def __getEmailCCID(self, email):
        """Get the CCID of the corresponding email"""

        email_resp = requests.get(
            f"{self.authzsvc_endpoint}Identity/by_email/{email}", headers={"Authorization": f"Bearer {self.api_token}"}
        )
        email_resp.raise_for_status()
        return email_resp.json()

    def __setRoleCCID(self, group_id, ccid, role):
        """Set CCID in eGroup."""

        lastShifterList = []
        lastShifterDelete = {}
        lastShifterDeleteUPN = []

        self.log.info("Getting current group identity members...")
        members_resp = requests.get(
            f"{self.authzsvc_endpoint}Group/{group_id}/members/identities",
            headers={"Authorization": f"Bearer {self.api_token}"},
        )
        members_resp.raise_for_status()
        for i in members_resp.json()["data"]:
            lastShifterList.append(i["personId"])
            lastShifterDelete[i["personId"]] = i["upn"]

        for lastShifter in lastShifterList:
            if lastShifter not in ccid:
                self.log.info(f"{lastShifter} is not anymore shifter, deleting ...")
                lastShifterDeleteUPN.append(lastShifterDelete[lastShifter])

        self.__removeMember(lastShifterDeleteUPN, group_id)

        # Adding a member means it will be the only one in the eGroup, as it is overwritten
        if ccid:
            res = self.__addMember(ccid, group_id, role)
            self.log.info(ccid)
            if not res["OK"]:
                self.log.error(res["Message"])
                return res
            self.log.info(f"{ccid} added successfully to the eGroup for role {role}")

        self.newShifters[role] = group_id

        return S_OK()

    def __removeMember(self, upns, group_id):
        self.log.info("Removing", f"UPNs {upns} to group {group_id}...")
        del_resp = requests.delete(
            f"{self.authzsvc_endpoint}Group/{group_id}/members/identities/",
            headers={"Authorization": f"Bearer {self.api_token}", "Content-Type": "application/json"},
            json=upns,
        )
        if del_resp.ok:
            self.log.info("Members deleted successfully.")
            for identity in del_resp.json()["data"]:
                self.log.verbose(identity["memberIdentity"]["id"], " has been removed")
        else:
            message = "Failed to remove members."
            if del_resp.status_code == 400:
                self.log.error(f"{message} No member to delete")
            else:
                self.log.error(message, f"{del_resp.status_code}: {del_resp.text}")

    def __addMember(self, ccid, group_id, role):
        """Adds a new member to the group."""

        self.log.info(f"Adding member {ccid} to Group {group_id} for role {role}")

        post_resp = requests.post(
            f"{self.authzsvc_endpoint}Group/{group_id}/members/identities/",
            headers={"Authorization": f"Bearer {self.api_token}", "Content-Type": "application/json"},
            json=ccid,
        )

        if not post_resp.ok:
            self.log.error("Failed to add members.", f"{post_resp.status_code}: {post_resp.text}")
            return S_ERROR(post_resp.text)

        self.log.info("Members added successfully.")
        for identity in post_resp.json()["data"]:
            self.log.verbose(identity["identityId"], " has been added")
        return S_OK()

    def __notifyNewShifter(self, role, eGroup):
        """Sends an email to the shifter ( if any ) at the beginning of the shift
        period."""

        if role != "Production":
            self.log.info(f"No email body defined for {role} role")
            return S_OK()
        body = __productionBody__

        prodRole = self.roles["Production"]
        geocRole = self.roles["Grid Expert"]
        body = body % (self.roleShifters[prodRole][0], self.roleShifters[geocRole][0])

        # Hardcoded Concezio's email to avoid dirac@mail.cern.ch be rejected by smtp server
        res = self.diracAdmin.sendMail(
            f"{eGroup}@cern.ch", "Shifter information", body, fromAddress="concezio.bozzi@cern.ch"
        )
        return res


__productionBody__ = """Dear GEOC,

this is an (automatic) mail to welcome you on the grid operations shifts.
In order to facilitate your shift activities we wanted to provide you some pointers,
where you could find more information about shifts, the activities and your duties during this period.

http://lhcb-shifters.web.cern.ch/


LHCbDIRAC portal
https://lhcb-portal-dirac.cern.ch/DIRAC


The logbook for LHCb operations, where all activities concerning offline operation are being logged.
https://lblogbook.cern.ch/Operations/
"""



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2025 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Reproduce DIRAC jobs locally for debugging and testing purposes.

This script helps you download job logs and execute jobs locally using the same
environment and inputs as the original job on the grid.
"""
from __future__ import annotations

import os
import re
import subprocess
import sys
import tempfile
from pathlib import Path
from typing import Optional

import typer
from rich.console import Console
from rich.panel import Panel
from typing import Annotated

import DIRAC

app = typer.Typer(
    name="dirac-wms-job-reproduce",
    help="Reproduce DIRAC jobs locally for debugging and testing",
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
)
console = Console()


def parse_identifier(identifier: str) -> tuple[str, str | int]:
    """Parse the input identifier and determine its type.

    Returns:
        tuple of (identifier_type, value) where identifier_type is one of:
        - "job_id": integer job ID
        - "logse_url": root:// URL to LogSE
        - "logse_lfn": LFN path to log file
    """
    # Check if it's a LogSE URL
    if identifier.startswith("root://"):
        return ("logse_url", identifier)

    # Check if it's a LogSE LFN (with or without LFN: prefix)
    # Format: /lhcb/.../LOG/... or LFN:/lhcb/.../LOG/...
    lfn = identifier.removeprefix("LFN:")
    if lfn.startswith("/lhcb/") and "/LOG/" in lfn:
        return ("logse_lfn", lfn)

    # Try to parse as job ID
    try:
        job_id = int(identifier)
        return ("job_id", job_id)
    except ValueError:
        console.print(f"[red]Could not parse identifier: {identifier}[/red]")
        console.print("[yellow]Expected one of:[/yellow]")
        console.print("  - Job ID: 1234567")
        console.print("  - LogSE URL: root://eoslhcb.cern.ch//eos/lhcb/grid/prod/lhcb/logSE/...")
        console.print("  - LogSE LFN: /lhcb/LHCb/Collision25/LOG/00329360/0000/00001030.zip")
        console.print("  - LogSE LFN: LFN:/lhcb/LHCb/Collision25/LOG/00329360/0000/00001030.zip")
        raise typer.Exit(1)


def get_log_path_from_logse_url(logse_url: str) -> str:
    """Extract the log path from a LogSE URL.

    Example input: root://eoslhcb.cern.ch//eos/lhcb/grid/prod/lhcb/logSE/lhcb/LHCb/Collision25/LOG/00329360/0000/00001030.zip
    Returns: lhcb/LHCb/Collision25/LOG/00329360/0000/00001030
    """
    # Extract path after logSE/
    match = re.search(r"logSE/(.+?)(\.zip)?$", logse_url)
    if not match:
        console.print(f"[red]Could not parse LogSE URL: {logse_url}[/red]")
        raise typer.Exit(1)

    log_path = match.group(1)
    console.print(f"[green]Extracted log path:[/green] {log_path}")
    return log_path


def get_log_path_from_logse_lfn(logse_lfn: str) -> str:
    """Extract the log path from a LogSE LFN.

    Example input: /lhcb/LHCb/Collision25/LOG/00329360/0000/00001030.zip
    Returns: lhcb/LHCb/Collision25/LOG/00329360/0000/00001030
    """
    # Remove leading slash and .zip extension if present
    log_path = logse_lfn.lstrip("/")
    if log_path.endswith(".zip"):
        log_path = log_path[:-4]

    console.print(f"[green]Extracted log path:[/green] {log_path}")
    return log_path


def get_log_path_from_job_id(job_id: int) -> str:
    """Get the EOS log path for a given job ID.

    Queries DIRAC for the job parameters to find the log URL, then extracts
    the EOS path from it.
    """
    console.print(f"[blue]Fetching job parameters for job {job_id}...[/blue]")

    from DIRAC.Interfaces.API.Dirac import Dirac

    dirac = Dirac()
    result = dirac.getJobParameters(job_id)

    if not result["OK"]:
        console.print(f"[red]Failed to get job parameters: {result['Message']}[/red]")
        raise typer.Exit(1)

    params = result["Value"]

    # Extract log path from 'Log URL' parameter
    # Format: '<a href="https://lhcb-dirac-logse.web.cern.ch:443/lhcb/MC/2024/LOG/00335034/0000/00000030/">Log file directory</a>'
    log_url_param = params.get("Log URL", "")

    url_match = re.search(r'href="([^"]+)"', log_url_param)
    if not url_match:
        console.print(f"[red]No Log URL found in job parameters[/red]")
        console.print(f"[dim]Available params: {list(params.keys())}[/dim]")
        raise typer.Exit(1)

    log_url = url_match.group(1)

    # Extract path from URL (after the hostname)
    # URL format: https://lhcb-dirac-logse.web.cern.ch:443/lhcb/MC/2024/LOG/...
    path_match = re.search(r"https?://[^/]+(/lhcb/.+?)/?$", log_url)
    if not path_match:
        console.print(f"[red]Could not extract path from Log URL: {log_url}[/red]")
        raise typer.Exit(1)

    log_path = path_match.group(1).lstrip("/")
    console.print(f"[green]Found log path:[/green] {log_path}")

    return log_path


@app.command(name="download")
def download_job(
    identifier: Annotated[
        str,
        typer.Argument(help="Job ID, LogSE URL, or LogSE LFN"),
    ],
    output_dir: Annotated[
        Path | None,
        typer.Option("--output-dir", "-o", help="Directory to download job files to (default: creates temp dir)"),
    ] = None,
    keep_zip: Annotated[bool, typer.Option("--keep-zip", help="Keep the downloaded zip file")] = False,
):
    """Download job logs and set up reproducer environment.

    Accepts three identifier formats:
    - Job ID: 1234567
    - LogSE URL: root://eoslhcb.cern.ch//eos/lhcb/grid/prod/lhcb/logSE/...
    - LogSE LFN: /lhcb/LHCb/Collision25/LOG/00329360/0000/00001030.zip

    This command:
    1. Resolves the identifier to get the job's log location
    2. Downloads the job logs from EOS
    3. Extracts the logs to a directory

    You can then run the job locally using the 'run' command.
    """

    # Check for conflicting environment variables that would interfere with job execution
    conflicting_vars = []
    if "JOBID" in os.environ:
        conflicting_vars.append(("JOBID", os.environ["JOBID"]))
    if "DIRACJOBID" in os.environ:
        conflicting_vars.append(("DIRACJOBID", os.environ["DIRACJOBID"]))

    if conflicting_vars:
        console.print()
        console.print(f"[red]{'🐲 Dragons! ' * 8}[/red]")
        console.print()
        console.print("[red]Error: Conflicting environment variables detected[/red]")
        console.print()
        console.print("The following environment variables are set and will cause nasty things to happen:")
        for var_name, var_value in conflicting_vars:
            console.print(f"  [yellow]{var_name}[/yellow] = {var_value}")
        console.print()
        console.print("Please unset these variables before running this script:")
        for var_name, _ in conflicting_vars:
            console.print(f"  [cyan]unset {var_name}[/cyan]")
        console.print()
        console.print(f"[red]{'🐲 Dragons! ' * 8}[/red]")
        console.print()
        raise typer.Exit(1)

    DIRAC.initialize()

    # Parse the identifier
    id_type, value = parse_identifier(identifier)

    # Resolve to job ID and log path
    if id_type == "job_id":
        job_id = value
        log_path = get_log_path_from_job_id(job_id)
    elif id_type == "logse_url":
        log_path = get_log_path_from_logse_url(value)
        # We don't have a job ID in this case, just use the log path for identification
        job_id = None
    elif id_type == "logse_lfn":
        log_path = get_log_path_from_logse_lfn(value)
        # We don't have a job ID in this case, just use the log path for identification
        job_id = None
    else:
        console.print(f"[red]Unknown identifier type: {id_type}[/red]")
        raise typer.Exit(1)

    # Create output directory
    if output_dir is None:
        if job_id:
            output_dir = Path(tempfile.mkdtemp(prefix=f"job_{job_id}_"))
        else:
            output_dir = Path(tempfile.mkdtemp(prefix="job_"))
        console.print(f"[blue]Created temporary directory:[/blue] {output_dir}")
    else:
        output_dir = Path(output_dir).resolve()
        output_dir.mkdir(parents=True, exist_ok=True)
        console.print(f"[blue]Using output directory:[/blue] {output_dir}")

    # Use log path basename for zip filename if we don't have a job ID
    if job_id:
        zip_file = output_dir / f"{job_id}.zip"
    else:
        zip_basename = Path(log_path).name
        zip_file = output_dir / f"{zip_basename}.zip"

    # Download with xrdcp
    xrd_url = f"root://eoslhcb.cern.ch//eos/lhcb/grid/prod/lhcb/logSE/{log_path}.zip"
    console.print(f"[blue]Downloading logs from EOS...[/blue]")
    console.print(f"[dim]Source: {xrd_url}[/dim]")

    try:
        subprocess.run(
            ["xrdcp", xrd_url, str(zip_file)],
            check=True,
            capture_output=True,
        )
    except subprocess.CalledProcessError as e:
        console.print(f"[red]Failed to download logs:[/red]")
        console.print(e.stderr.decode())
        raise typer.Exit(1)
    except FileNotFoundError:
        console.print("[red]xrdcp command not found. Please ensure xrootd client is installed.[/red]")
        raise typer.Exit(1)

    console.print(f"[green]Downloaded to {zip_file}[/green]")

    # Unzip
    console.print(f"[blue]Extracting logs...[/blue]")
    try:
        subprocess.run(
            ["unzip", "-q", str(zip_file), "-d", str(output_dir)],
            check=True,
            capture_output=True,
        )
    except subprocess.CalledProcessError as e:
        console.print(f"[red]Failed to extract logs:[/red]")
        console.print(e.stderr.decode())
        raise typer.Exit(1)

    # Find the extracted directory (usually there's one subdirectory)
    subdirs = [d for d in output_dir.iterdir() if d.is_dir()]
    if len(subdirs) == 1:
        job_dir = subdirs[0]
    else:
        # Multiple or no subdirectories, use output_dir
        job_dir = output_dir

    console.print(f"[green]Logs extracted to {job_dir}[/green]")

    # Remove zip file unless requested to keep it
    if not keep_zip:
        zip_file.unlink()
        console.print(f"[dim]Removed {zip_file}[/dim]")

    # Display next steps
    console.print()
    if job_id:
        title_msg = f"[green]Job logs downloaded successfully![/green]\n\n[blue]Job ID:[/blue] {job_id}\n"
    else:
        title_msg = f"[green]Job logs downloaded successfully![/green]\n\n"

    console.print(
        Panel.fit(
            f"{title_msg}"
            f"[blue]Job directory:[/blue] {job_dir}\n\n"
            f"[yellow]Next steps:[/yellow]\n"
            f"  1. Review the job files in {job_dir}\n"
            f"  2. Run the job locally:\n"
            f"     [cyan]dirac-wms-job-reproduce run {job_dir}[/cyan]\n"
            f"     or: [cyan]cd {job_dir} && (lb-dirac) dirac-jobexec jobDescription.xml[/cyan]\n"
            f"  3. Create a minimal reproducer if needed for bug reports\n\n",
            title="Download Complete",
            border_style="green",
        )
    )


@app.command(name="run")
def run_reproducer(
    job_dir: Annotated[Path, typer.Argument(help="Job directory (from 'download' command)")],
    dry_run: Annotated[bool, typer.Option("--dry-run", help="Show commands without executing")] = False,
    download: Annotated[
        bool, typer.Option("--download", help="Download input files (sets inputDataPolicy to DownloadInputData)")
    ] = False,
    protocol: Annotated[
        bool, typer.Option("--protocol", help="Access files by protocol (sets inputDataPolicy to InputDataByProtocol)")
    ] = False,
):
    """Run the reproducer from a downloaded job directory.

    This command:
    1. Extracts input data LFNs from jobDescription.xml
    2. Generates the XML catalog with bookkeeping information
    3. Executes the job locally using dirac-jobexec

    The job will run in the same environment as on the grid, using the original
    jobDescription.xml and input files.

    By default, uses InputDataByProtocol. Use --download to download files locally first,
    or --protocol to explicitly use protocol-based access.
    """
    # Validate flags
    if download and protocol:
        console.print("[red]Error: Cannot specify both --download and --protocol[/red]")
        raise typer.Exit(1)

    # Check for conflicting environment variables that would interfere with job execution
    conflicting_vars = []
    if "JOBID" in os.environ:
        conflicting_vars.append(("JOBID", os.environ["JOBID"]))
    if "DIRACJOBID" in os.environ:
        conflicting_vars.append(("DIRACJOBID", os.environ["DIRACJOBID"]))

    if conflicting_vars:
        console.print()
        console.print(f"[red]{'🐲 Dragons! ' * 8}[/red]")
        console.print()
        console.print("[red]Error: Conflicting environment variables detected[/red]")
        console.print()
        console.print("The following environment variables are set and will cause nasty things to happen:")
        for var_name, var_value in conflicting_vars:
            console.print(f"  [yellow]{var_name}[/yellow] = {var_value}")
        console.print()
        console.print("Please unset these variables before running this script:")
        for var_name, _ in conflicting_vars:
            console.print(f"  [cyan]unset {var_name}[/cyan]")
        console.print()
        console.print(f"[red]{'🐲 Dragons! ' * 8}[/red]")
        console.print()
        raise typer.Exit(1)

    # Determine input data policy
    if download:
        input_data_policy = "DIRAC.WorkloadManagementSystem.Client.DownloadInputData"
    else:
        # Default to protocol-based access
        input_data_policy = "DIRAC.WorkloadManagementSystem.Client.InputDataByProtocol"

    DIRAC.initialize()

    job_dir = Path(job_dir).resolve()

    if not job_dir.exists():
        console.print(f"[red]Job directory does not exist: {job_dir}[/red]")
        raise typer.Exit(1)

    # Check for jobDescription.xml
    job_desc = job_dir / "jobDescription.xml"
    if not job_desc.exists():
        console.print(f"[red]jobDescription.xml not found in {job_dir}[/red]")
        raise typer.Exit(1)

    console.print(f"[blue]Working directory:[/blue] {job_dir}")
    console.print(f"[blue]Job description:[/blue] {job_desc}")

    # Extract input data LFNs and generate XML catalog
    console.print(f"[blue]Extracting input data LFNs from jobDescription.xml...[/blue]")

    try:
        # Read jobDescription.xml and extract LFNs
        with open(job_desc) as f:
            job_xml = f.read()

        # Find InputData parameter (handles both CDATA and plain text)
        # Format 1: <![CDATA[/lhcb/.../file1.dst;/lhcb/.../file2.dst;...]]>
        # Format 2: <value>/lhcb/.../file1.dst;/lhcb/.../file2.dst;...</value>
        lfn_match = re.search(
            r'name="InputData"[^>]*>.*?<!\[CDATA\[(.*?)\]\]>|name="InputData"[^>]*><value>(.*?)</value>',
            job_xml,
            re.DOTALL,
        )

        if lfn_match:
            input_data = lfn_match.group(1) or lfn_match.group(2) or ""
            # Split by semicolons - handles format like:
            # /lhcb/.../file1.dst;/lhcb/.../file2.dst;...
            lfns = [lfn.strip() for lfn in input_data.split(";") if lfn.strip() and lfn.strip().startswith("/lhcb")]

            if lfns:
                console.print(f"[green]Found {len(lfns)} input file(s)[/green]")
                for lfn in lfns[:5]:  # Show first 5
                    console.print(f"  [dim]{lfn}[/dim]")
                if len(lfns) > 5:
                    console.print(f"  [dim]... and {len(lfns) - 5} more[/dim]")
            else:
                console.print("[yellow]No input files found in job description[/yellow]")
                lfns = []
        else:
            console.print("[yellow]No InputData parameter found in job description[/yellow]")
            lfns = []

    except Exception as e:
        console.print(f"[red]Failed to parse jobDescription.xml: {e}[/red]")
        raise typer.Exit(1)

    # Generate commands
    commands = []

    if lfns:
        # Generate XML catalog
        policy_name = input_data_policy.split(".")[-1]
        console.print(f"[blue]Generating XML catalog from {len(lfns)} LFN(s)...[/blue]")
        console.print(f"[dim]Input data policy: {policy_name}[/dim]")

        from DIRAC.Interfaces.API.Dirac import Dirac

        dirac = Dirac()
        catalog_result = dirac.getInputDataCatalog(lfns, [], "pool_xml_catalog.xml", inputDataPolicy=input_data_policy)

        if not catalog_result["OK"]:
            console.print(f"[red]Failed to generate XML catalog: {catalog_result['Message']}[/red]")
            raise typer.Exit(1)

        print(catalog_result)

        console.print(f"[green]✓ XML catalog generated[/green]")

    # Execute job
    console.print(f"[blue]Preparing to execute job...[/blue]")

    jobexec_cmd = [
        "dirac-jobexec",
        str(job_desc),
    ]

    commands.append(("Execute job", jobexec_cmd, None))

    # Show or execute commands
    if dry_run:
        console.print()
        console.print("[yellow]Dry run mode - commands that would be executed:[/yellow]")
        console.print()

        for i, (desc, cmd, stdin_data) in enumerate(commands, 1):
            console.print(f"[cyan]{i}. {desc}:[/cyan]")
            if stdin_data:
                console.print(f"   echo {repr(stdin_data)} | {' '.join(cmd)}")
            else:
                console.print(f"   {' '.join(cmd)}")
            console.print()

        console.print("[yellow]Run without --dry-run to execute[/yellow]")
    else:
        # Execute commands
        console.print()
        console.print("[green]Executing reproducer...[/green]")
        console.print()

        for desc, cmd, stdin_data in commands:
            console.print(f"[cyan]Running: {desc}[/cyan]")
            console.print(f"[dim]Command: {' '.join(cmd)}[/dim]")

            try:
                if stdin_data:
                    result = subprocess.run(
                        cmd,
                        input=stdin_data,
                        text=True,
                        cwd=job_dir,
                        check=True,
                    )
                else:
                    result = subprocess.run(
                        cmd,
                        cwd=job_dir,
                        check=True,
                    )

                console.print(f"[green]✓ {desc} completed[/green]")
            except subprocess.CalledProcessError as e:
                console.print(f"[red]✗ {desc} failed with exit code {e.returncode}[/red]")
                raise typer.Exit(1)
            except FileNotFoundError:
                console.print(f"[red]Command not found: {cmd[0]}[/red]")
                raise typer.Exit(1)

            console.print()

        console.print("[green]Job execution completed![/green]")


@app.command(name="reproduce")
def reproduce_job(
    identifier: Annotated[
        str,
        typer.Argument(help="Job ID, LogSE URL, or LogSE LFN"),
    ],
    output_dir: Annotated[
        Path | None,
        typer.Option("--output-dir", "-o", help="Directory to download job files to (default: creates temp dir)"),
    ] = None,
    keep_zip: Annotated[bool, typer.Option("--keep-zip", help="Keep the downloaded zip file")] = False,
    dry_run: Annotated[bool, typer.Option("--dry-run", help="Show commands without executing")] = False,
    download: Annotated[
        bool, typer.Option("--download", help="Download input files (sets inputDataPolicy to DownloadInputData)")
    ] = False,
    protocol: Annotated[
        bool, typer.Option("--protocol", help="Access files by protocol (sets inputDataPolicy to InputDataByProtocol)")
    ] = False,
):
    """Download and run a job reproducer in one step.

    Accepts three identifier formats:
    - Job ID: 1234567
    - LogSE URL: root://eoslhcb.cern.ch//eos/lhcb/grid/prod/lhcb/logSE/...
    - LogSE LFN: /lhcb/LHCb/Collision25/LOG/00329360/0000/00001030.zip

    This is a convenience command that combines 'download' and 'run'.
    Equivalent to running:
      dirac-wms-job-reproduce download <IDENTIFIER> -o <OUTPUT_DIR>
      dirac-wms-job-reproduce run <OUTPUT_DIR> [--download|--protocol]

    By default, uses InputDataByProtocol. Use --download to download files locally first,
    or --protocol to explicitly use protocol-based access.
    """
    # Download the job
    download_job(identifier, output_dir, keep_zip)

    # Find the job directory
    if output_dir is None:
        console.print("[red]Cannot automatically determine job directory from temp dir[/red]")
        console.print("[yellow]Please use 'download' and 'run' commands separately[/yellow]")
        raise typer.Exit(1)

    output_dir = Path(output_dir).resolve()
    subdirs = [d for d in output_dir.iterdir() if d.is_dir()]

    if len(subdirs) == 1:
        job_dir = subdirs[0]
    else:
        job_dir = output_dir

    # Add a separator
    console.print()
    console.rule("[bold]Starting Job Execution[/bold]")
    console.print()

    # Run the reproducer
    run_reproducer(job_dir, dry_run, download, protocol)


def main():
    """Entry point for the CLI."""
    app()


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2025 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import logging
from threading import Lock
from urllib.parse import quote_plus

from cachetools import TTLCache, cached
import requests

from DIRAC import gConfig, gLogger

from LHCbDIRAC.ConfigurationSystem.Utilities.CERNSSOAPIToken import CERNSSOAPIToken


class FenceIntegration:
    """Class for talking to the LHCb Fence membership API."""

    def __init__(self, fence_url=None, token: CERNSSOAPIToken | None = None):
        """Initialize FenceIntegration.

        Args:
            fence_url: Fence API base URL. If None, will try to get from DIRAC config.
            token: CERNSSOAPIToken object. If None, will try to set-up from DIRAC config.
        """
        self.fence_url = fence_url or self._get_fence_url()
        self.token = token or self._get_token()
        self.log = gLogger.getSubLogger("FenceIntegration")

    def _get_fence_url(self):
        """Get Fence URL from DIRAC configuration."""
        return gConfig.getValue(
            "/Systems/Configuration/Agents/VOMS2CSAgent/FenceIntegration/URL", "https://lbfence.cern.ch"
        )

    def _get_token(self) -> CERNSSOAPIToken:
        """Get exchange-able token from DIRAC configuration."""
        audience = gConfig.getValue(
            "/Systems/Configuration/Agents/VOMS2CSAgent/FenceIntegration/Audience", "lhcb-membership-api-prod"
        )
        client_id = gConfig.getValue(
            "/Systems/Configuration/Agents/VOMS2CSAgent/FenceIntegration/ClientID", "lhcbdirac-fence-integration"
        )
        client_secret = gConfig.getValue("/Systems/Configuration/Agents/VOMS2CSAgent/FenceIntegration/ClientSecret", "")
        return CERNSSOAPIToken(audience=audience, client_id=client_id, client_secret=client_secret)

    def membership_search(self, query, *, offset: int = 0, limit: int = 1000):
        """Search for members in Fence.

        Args:
            query: Search query string
            offset: Result offset for pagination
            limit: Maximum number of results to return

        Returns:
            Tuple of (total_count, results_list)
        """
        data = {"results": []}
        try:
            result = requests.get(
                (
                    f"{self.fence_url}/membership/api/members/search"
                    f"?offset={offset}&limit={limit}&queryString={quote_plus(query)}"
                ),
                headers={
                    "Accept": "application/json",
                    "Authorization": f"Bearer {self.token.token}",
                },
                verify=False,
                timeout=45,
            )
            result.raise_for_status()
            data = result.json()
        except Exception as e:
            self.log.error("Failed to query fence", str(e))
            raise

        return int(data.get("numberOfResults", 0)), data.get("results", [])

    def get_members(self, query_string):
        """Get all members matching a query string.

        Args:
            query_string: Fence query string to search for members

        Returns:
            List of member dictionaries
        """
        member_results = []
        n_processed = 0
        n_members = float("inf")  # Initialize to infinity so the loop runs at least once
        page = 0
        limit = 500
        self.log.info("Querying fence with:", query_string)
        while n_processed < n_members:
            self.log.info("Querying fence for members:", f"offset={page * limit}, limit={limit}")
            n_members, members = self.membership_search(query=query_string, offset=page * limit, limit=limit)
            member_results.extend(members)
            n_processed += len(members)
            page += 1
            self.log.info("Processed members from fence", f"{n_processed}/{n_members}")

        self.log.info("Finished querying members from fence", f"{n_processed} total")
        return member_results

    def get_all_memberships(self):
        """Query all fence memberships across all employment statuses.

        Returns:
            List of all member dictionaries
        """
        query_strings = [
            """( "employmentStatus" = "Inactive" )""",
            """( "employmentStatus" = "Active" )""",
            """( "employmentStatus" = "Extended" )""",
            """( "employmentStatus" = "Upcoming" )""",
        ]
        everyone_agg = []
        for query_string in query_strings:
            everyone_agg.extend(self.get_members(query_string=query_string))
        return everyone_agg



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
.. versionadded:: v11.0.53

This agent consolidates data from the bookkeeping and the DFC, generates json out of it,
uploads them to a storage.
It also sends the accounting data to OpenSearch, and fill in 2 tables in the StorageUsageDB

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN UserStorageDumpAgent
  :end-before: ##END UserStorageDumpAgent
  :dedent: 2
  :caption: UserStorageDumpAgent options

"""
# # imports
import asyncio
import json
import re
from collections import defaultdict
from datetime import datetime, timezone, timedelta
from functools import lru_cache, partial
from pathlib import Path
from textwrap import dedent
from urllib.parse import quote_plus

from opensearchpy import OpenSearch, helpers


import numpy as np
import pandas as pd
import time


# # from DIRAC
from DIRAC import S_OK, S_ERROR, gLogger, gConfig
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise
from DIRAC.ConfigurationSystem.Client.Utilities import getDBParameters
from DIRAC.DataManagementSystem.DB.FileCatalogDB import FileCatalogDB
from LHCbDIRAC.DataManagementSystem.DB.StorageUsageDumpDB import StorageUsageDumpDB


# Before 90 days, we keep all the dumps
# After 90 days, we keep the first per day
# After a year, we keep the first per week
OS_DAILY_RETENTION = 90
OS_WEEKLY_RETENTION = 365

# We keep all the weekly dumps
# We keep the daily dump for a year
S3_DAILY_RETENTION = 365

AGENT_NAME = "DataManagement/UserStorageDumpAgent"

USER_ROOT_DIR = "/lhcb/user"

S3_BUCKET_NAME = "lhcbdirac"
S3_ENDPOINT_URL = "https://s3.cern.ch:443"
OS_ENDPOINT_URL = "os-lhcb-dirac.cern.ch:443/os"
OS_INDEX_PREFIX = "lhcb_storage_user_usage_index_"


async def returnValueOrRaiseAsync(func, *args, **kwargs):
    loop = asyncio.get_running_loop()
    return returnValueOrRaise(await loop.run_in_executor(None, func, *args, **kwargs))


def sendToOpenSearch(df: pd.DataFrame, hosts: str, http_auth: tuple[str, str]):
    """
    Sends the data to OpenSearch
    """
    log = gLogger.getSubLogger("sendToOpenSearch")
    log.info("Sending to OpenSearch")
    start_time = time.time()

    fields = {
        "SEName": {"type": "keyword"},
        "User": {"type": "keyword"},
        "UserStatus": {"type": "keyword"},
        "SESize": {"type": "long"},
        "SEFiles": {"type": "long"},
        "timestamp": {"type": "date"},
    }

    # Create an OpenSearch client
    client = OpenSearch(
        hosts=hosts,
        http_auth=http_auth,
        use_ssl=True,
        verify_certs=False,
        ssl_show_warn=False,
    )

    template_body = {
        "template": {"mappings": {"properties": fields}},
        "index_patterns": [f"{OS_INDEX_PREFIX}*"],
        "priority": 5,
    }
    result = client.indices.put_index_template(name=OS_INDEX_PREFIX, body=template_body)
    assert result["acknowledged"]

    utc_now = datetime.now(tz=timezone.utc)
    timestamp = utc_now.strftime("%Y-%m-%dT%H:%M")
    as_date = utc_now.isocalendar()
    index_timestamp = f"{as_date[0]}-{as_date[1]}"

    data = df.reset_index(drop=True).to_dict("records")

    actual_index = f"{OS_INDEX_PREFIX}{index_timestamp}"
    for entry in data:
        # entry.pop("Unnamed: 0")

        entry["_index"] = actual_index
        entry["_id"] = f"{entry['SEName']}_{entry['User']}_{timestamp}"
        entry["timestamp"] = utc_now

    succeeded = []
    failed = []
    for success, item in helpers.parallel_bulk(
        client,
        actions=data,
        chunk_size=500,
        raise_on_error=False,
        raise_on_exception=False,
        max_chunk_bytes=20 * 1024 * 1024,
        request_timeout=60,
    ):

        if success:
            succeeded.append(item)
        else:
            failed.append(item)

    if len(failed) > 0:
        log.error("There were errors: ", f"{len(failed)}")
        for item in failed:
            log.error(f"    {item}")

    if len(succeeded) > 0:
        log.info("Bulk-inserted", f"{len(succeeded)} items.")
    log.info("Upload duration", f"{time.time() - start_time:.2f} secs")
    return S_OK()


async def get_users_usage():
    db = FileCatalogDB()
    columns = ["Name", "SEName", "SESize", "SEFiles"]
    parent_id = await returnValueOrRaiseAsync(
        db._query,
        f"SELECT DirID from FC_DirectoryList where Name = '{USER_ROOT_DIR}'",
    )

    query = """
    with parent_id as (select ChildID from FC_DirectoryClosure where ParentID = %s and Depth = 2)
    SELECT d.Name, SEName, SUM(du.SESize), SUM(du.SEFiles)
    FROM FC_DirectoryUsage du
    JOIN FC_DirectoryClosure dc ON du.DirID = dc.ChildID
    JOIN parent_id p ON dc.ParentID = p.ChildID
    JOIN FC_DirectoryList d on d.DirID = p.ChildID
    JOIN FC_StorageElements se on se.SEID = du.SEID
    group by p.ChildID, du.SEID;
    """

    # We can't work with db._query because we change the isolation level
    conn = returnValueOrRaise(db._getConnection())

    with conn.cursor() as cursor:
        cursor.execute("SET SESSION TRANSACTION ISOLATION LEVEL READ UNCOMMITTED;")

        cursor.execute("SELECT @@transaction_isolation;")

        query = dedent(query.strip())
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, cursor.execute, query, parent_id)
        return columns, await loop.run_in_executor(None, cursor.fetchall)


def get_user_status(username: str, users_status: dict[str:bool]):
    if username not in users_status:
        return "Missing"
    if users_status[username]:
        return "Suspended"
    return "OK"


class UserStorageDumpAgent(AgentModule):

    def initialize(self):
        self.workDirectory = self.am_getWorkDirectory()
        self.os_endpoint_url = self.am_getOption("OS_endpoint_url", OS_ENDPOINT_URL)
        self.os_username = gConfig.getValue("/Systems/NoSQLDatabases/User")
        self.os_password = gConfig.getValue("/Systems/NoSQLDatabases/Password")

        self.storageUsageDumpDB = StorageUsageDumpDB()

        return S_OK()

    def execute(self):
        asyncio.run(self.do_the_dump())
        return S_OK()

    async def do_the_dump(self):

        user_storage = await get_users_usage()
        dfc_df = pd.DataFrame(user_storage[1], columns=user_storage[0])
        dfc_df["User"] = dfc_df.Name.apply(lambda x: x.split("/")[-1])

        users_status = {
            user: ("lhcb" in gConfig.getOptionsDict(f"/Registry/Users/{user}")["Value"].get("Suspended", ""))
            for user in gConfig.getSections("/Registry/Users")["Value"]
        }
        _get_user_status = partial(get_user_status, users_status=users_status)
        dfc_df["UserStatus"] = dfc_df.User.apply(_get_user_status)
        dfc_df.drop("Name", axis=1, inplace=True)

        # Only keep the rows for which there are size and files
        dfc_df = dfc_df[(dfc_df.SESize != 0) & (dfc_df.SEFiles != 0)]

        dfc_df.to_csv("/tmp/user_storage.csv.zst", index=False)

        await asyncio.gather(
            returnValueOrRaiseAsync(
                partial(sendToOpenSearch, dfc_df, self.os_endpoint_url, (self.os_username, self.os_password))
            ),
            returnValueOrRaiseAsync(user_storage_to_sql, dfc_df, self.storageUsageDumpDB),
            # returnValueOrRaiseAsync(
            #     partial(
            #         storage_summary_to_sql,
            #         df,
            #         self.storageUsageDumpDB,
            #     )
            # ),
        )


def user_storage_to_sql(df, db: StorageUsageDumpDB):
    log = gLogger.getLocalSubLogger("user_storage_to_sql")
    log.info("Writing user directory metadata to database")
    start_time = time.time()
    db.user_storage_to_sql(df)
    log.info("Upload duration", f"{time.time() - start_time:.2f} secs")
    return S_OK()



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import annotations

from ..integrations import RepoIssueMetadataBase, RepoIssueBase, GitlabRepoBase


class APIssueMetadata(RepoIssueMetadataBase):
    # [discussion_id] = b64encode(zstd.compress(RequestFeedbackAction))
    open_discussions: dict[str, str] = {}


class APRepoIssue(RepoIssueBase, metadata_model=APIssueMetadata):
    pass


class APGitlabRepo(GitlabRepoBase, project_slug="lhcb-datapkg/AnalysisProductions"):
    pass



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import annotations

import asyncio
from collections import defaultdict
from collections.abc import Collection, Iterable
from datetime import datetime, timedelta
from functools import cached_property
from itertools import chain
from time import perf_counter
from typing import Any

import pytz
from rich import print
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.ProductionManagementSystem.Client.AnalysisProductionsClient import AnalysisProductionsClient
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequestClient import ProductionRequestClient
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

from .constants import ACTIVE_TRANSFORMATION_STATES
from ..LogAnalysis.utils import call_dirac
from .models import APTransformDebugInfo, APRequestDebugInfo, APDebugInfo, APBaseDebugInfo


async def get_active_productions():
    print("Getting active productions")
    start = perf_counter()
    result = await call_dirac(
        ProductionRequestClient().getProductionRequestList,
        0,  # Parent
        "RequestID",  # Sortby
        "DESC",  # Sort order
        0,  # Offset
        0,  # Max results
        {
            "RequestType": "AnalysisProduction",
            "RequestState": ",".join(["New", "Accepted", "Active", "PPG OK", "Submitted", "Tech OK"]),
        },
    )
    print(f"Got {result['Total']} active productions in {perf_counter() - start:.2f} seconds")
    return {x["RequestID"]: x for x in result["Rows"]}


async def get_ap_requests():
    print("Getting analysis production requests")
    start = perf_counter()
    result = await call_dirac(AnalysisProductionsClient().listRequests)
    print(f"Got {len(result)} analysis production requests in {perf_counter() - start:.2f} seconds")
    return result


async def get_archived_ap_requests():
    print("Getting archived analysis production requests")
    start = perf_counter()
    result = await call_dirac(AnalysisProductionsClient().getArchivedRequests)
    print(f"Got {len(result)} archived analysis production requests in {perf_counter() - start:.2f} seconds")
    return result


class Transforms:
    @classmethod
    async def from_transform_ids(cls, request_ids: Iterable[int], progress):
        print(f"Getting transformation info for {len(request_ids)} production requests")
        start = perf_counter()
        result = await call_dirac(
            TransformationClient().getTransformations,
            {"TransformationFamily": list(request_ids)},
            columns=[
                "TransformationID",
                "TransformationFamily",
                "Type",
                "Status",
                "GroupSize",
            ],
        )
        print(
            f"Got info for {len(result)} transforms from {len(request_ids)} requests in {perf_counter() - start:.2f} seconds"
        )
        for transform in result:
            transform["TransformationFamily"] = int(transform["TransformationFamily"])
        self = cls(result)
        await self._find_input_info()
        return self

    def __init__(self, transforms):
        self._by_id = {x["TransformationID"]: x for x in transforms}

    @cached_property
    def by_id(self):
        result = {}
        for transform_id, dict_data in self._by_id.items():
            result[transform_id] = APTransformDebugInfo(
                transform_id=transform_id,
                state=dict_data["Status"],
                type=dict_data["Type"],
                group_size=dict_data["GroupSize"],
                # file_counts=self._file_counts[transform_id],
                active_inputs=self._active_inputs[transform_id],
            )
        return result

    @cached_property
    def id_to_family(self):
        result = {}
        for transform_id, dict_data in self._by_id.items():
            result[transform_id] = dict_data["TransformationFamily"]
        return result

    @cached_property
    def by_family(self):
        result = defaultdict(list)
        for transform_id, dict_data in self._by_id.items():
            result[dict_data["TransformationFamily"]].append(transform_id)
        return {k: sorted(v) for k, v in result.items()}

    async def _find_input_info(self) -> None:
        print(f"Getting input queries for {len(self._by_id)} active productions")
        start = perf_counter()
        input_queries = await call_dirac(
            TransformationClient().getBookkeepingQueries,
            sorted(self._by_id),
        )
        print(f"Got input queries in {perf_counter() - start:.2f} seconds")
        result: dict[int, set[int]] = {}
        to_check = {}
        for tid, query in input_queries.items():
            if in_prods := query.get("ProductionID", []):
                if isinstance(in_prods, int):
                    in_prods = [in_prods]
                active_in_prods = {
                    in_prod
                    for in_prod in in_prods
                    if in_prod in self._by_id and self._by_id[in_prod]["Status"] in ACTIVE_TRANSFORMATION_STATES
                }
                if active_in_prods:
                    # The input transformation is active, no need to check the bookkeeping
                    result[tid] = active_in_prods
                    continue
            to_check[hash_dict(query)] = query

        async def get_input_productions(input_query):
            if "ProductionID" in input_query:
                return [input_query["ProductionID"]]
            result = await call_dirac(BookkeepingClient().getProductions, input_query)
            return [x[0] for x in result["Records"]]

        print(f"Getting production IDs for {len(to_check)} unique queries")
        start = perf_counter()
        async with asyncio.TaskGroup() as tg:
            tasks = {k: tg.create_task(get_input_productions(v)) for k, v in to_check.items()}
            input_ids_by_hash = {k: await v for k, v in tasks.items()}
            end = perf_counter()
            print(f"Got production IDs from queries in {end - start:.2f} seconds")

        input_ids = list(chain(*input_ids_by_hash.values()))
        print(f"Getting transform info for {len(input_ids)} IDs")
        start = perf_counter()
        transforms_info = await call_dirac(
            TransformationClient().getTransformations,
            condDict={"TransformationID": sorted(set(input_ids))},
            limit=10000,
            columns=["TransformationID", "TransformationFamily", "Status", "Type"],
        )
        transforms_info = {x["TransformationID"]: x for x in transforms_info}
        end = perf_counter()
        print(f"Got transform info in {end - start:.2f} seconds")

        for tid, query in input_queries.items():
            if tid in result:
                # The input transformation is an AP and we already checked it
                continue
            input_ids = input_ids_by_hash[hash_dict(query)]
            result[tid] = {
                input_tid
                for input_tid in input_ids_by_hash[hash_dict(query)]
                if input_tid > 0
                and transforms_info[input_tid]["Status"] in ACTIVE_TRANSFORMATION_STATES
                and not query.get("EndDate")
            }
        self._active_inputs = result

    async def find_file_counts(self, transform_ids: Collection[int], progress) -> None:
        ptask = progress.add_task("Getting file counts", total=len(transform_ids))

        async def get_files_count(tid):
            """In Python 3.13 this can be replaced with `async for task in as_completed...`"""
            return tid, await call_dirac(TransformationClient().getTransformationFilesCount, tid, "Status")

        async with asyncio.TaskGroup() as tg:
            tasks = [tg.create_task(get_files_count(tid)) for tid in transform_ids]
            for task in asyncio.as_completed(tasks):
                tid, file_counts = await task
                self.by_id[tid].file_counts = file_counts
                progress.update(ptask, advance=1)


def hash_dict(d: dict[str, Any]) -> int:
    items = []
    for k, v in d.items():
        if isinstance(v, dict):
            v = hash_dict(v)
        elif isinstance(v, list):
            v = frozenset(v)
        elif not isinstance(v, (int, str)):
            raise NotImplementedError(v, type(v))
        items.append((k, v))
    return hash(frozenset(items))


async def get_ap_info(wg, analysis, version) -> APDebugInfo:
    from LHCbDIRAC.ProductionManagementSystem.Client.AnalysisProductionsClient import (
        AnalysisProductionsClient,
    )

    aps = await call_dirac(
        AnalysisProductionsClient().getProductions,
        wg=wg,
        analysis=analysis,
        version=version,
        with_lfns=False,
        with_pfns=False,
        with_transformations=False,
    )
    urls = {ap.get("jira_task") for ap in aps}
    if len(urls) != 1:
        raise NotImplementedError(urls)
    validity_start = min(ap["validity_start"] for ap in aps)

    return APDebugInfo(
        wg=wg,
        analysis=analysis,
        version=version,
        created=validity_start.astimezone(pytz.timezone("Europe/Zurich")),
        issue_url=urls.pop(),
    )


async def get_raw_ap_infos(
    active_request_ids: Iterable[int], requests, progress
) -> dict[tuple[str, str, str] : APDebugInfo]:
    # Deduplicate the analyses
    keys = {
        (wg, analysis, request["version"])
        for request in requests
        for wg, analysis in request["analyses"]
        if request["request_id"] in active_request_ids
    }
    # Get the APDebugInfo objects in parallel
    result = {}
    ptask = progress.add_task("Getting AP info", total=len(keys))
    async with asyncio.TaskGroup() as tg:
        tasks = [tg.create_task(get_ap_info(wg, analysis, version)) for wg, analysis, version in keys]
        for task in asyncio.as_completed(tasks):
            ap = await task
            result[(ap.wg, ap.analysis, ap.version)] = ap
            progress.update(ptask, advance=1)
    return result


class LiveAPInfo:
    @classmethod
    async def load(cls, progress):
        async with asyncio.TaskGroup() as tg:
            active_productions = tg.create_task(get_active_productions())
            requests = tg.create_task(get_ap_requests())
            archived_requests = tg.create_task(get_archived_ap_requests())
            # active_productions = await active_productions
            transforms = tg.create_task(Transforms.from_transform_ids(await active_productions, progress))
            # requests = await requests
            raw_ap_infos = get_raw_ap_infos(await active_productions, await requests, progress)
        return cls(
            active_productions=await active_productions,
            requests=await requests,
            archived_requests=await archived_requests,
            transforms=await transforms,
            raw_ap_infos=await raw_ap_infos,
        )

    def __init__(
        self,
        active_productions,
        requests,
        archived_requests,
        transforms: Transforms,
        raw_ap_infos: dict[tuple[str, str, str], APDebugInfo],
    ):
        self._active_productions = active_productions
        self._requests = requests
        self._archived_requests = archived_requests
        self.transforms = transforms
        self._raw_ap_infos = raw_ap_infos
        self._handle_pending()

    @cached_property
    def request_id_to_ap_keys(self) -> dict[int, set[tuple[str, str, str]]]:
        return {
            request["request_id"]: {(wg, analysis, request["version"]) for wg, analysis in request["analyses"]}
            for request in self._requests
        }

    @cached_property
    def missing(self) -> set[int]:
        """If requests are missing in the AnalysisProductionsDB then something is wrong"""
        result = set(self._active_productions)
        result -= set(self.request_id_to_ap_keys)
        result -= {r.request_id for a in self.archived_requests for r in a.requests}
        return result

    def _handle_pending(self):
        recent: dict[tuple[str, str, str], APDebugInfo] = {}
        stuck: dict[tuple[str, str, str], APDebugInfo] = {}

        timezone = pytz.timezone("Europe/Zurich")
        now = datetime.now(timezone)
        for prod in self._active_productions.values():
            if prod["RequestState"] == "Active":
                continue
            time_since_created = now - prod["crTime"].astimezone(timezone)
            result = recent if time_since_created < timedelta(hours=6) else stuck
            request_id = prod["RequestID"]
            if request_id in self.missing:
                continue
            for key in self.request_id_to_ap_keys[request_id]:
                if key not in result:
                    result[key] = self._raw_ap_infos[key].model_copy(deep=True)
                request_info = APRequestDebugInfo(
                    request_id=request_id,
                    transforms=[self.transforms.by_id[x] for x in self.transforms.by_family.get(request_id, [])],
                )
                result[key].requests.append(request_info)

        for transform in self.transforms.by_id.values():
            if transform.state in {"PPG_OK", "New"}:
                request_id = self.transforms.id_to_family[transform.transform_id]
                for key in self.request_id_to_ap_keys[request_id]:
                    if key not in stuck:
                        stuck[key] = self._raw_ap_infos[key].model_copy(deep=True)
                    request_info = APRequestDebugInfo(
                        request_id=request_id,
                        transforms=[self.transforms.by_id[x] for x in self.transforms.by_family.get(request_id, [])],
                    )
                    stuck[key].requests.append(request_info)

        self.recent = sorted(recent.values(), key=lambda x: x.created)
        self.stuck = sorted(stuck.values(), key=lambda x: x.created)

    @cached_property
    def archived_requests(self) -> list[APBaseDebugInfo]:
        by_version = {}
        for request in self._archived_requests:
            request_id = request["request_id"]
            if request_id not in self._active_productions:
                continue
            if request["version"] not in by_version:
                by_version[request["version"]] = APBaseDebugInfo(
                    version=request["version"], issue_url=request["jira_task"]
                )
            request_info = APRequestDebugInfo(
                request_id=request_id,
                transforms=[self.transforms.by_id[x] for x in self.transforms.by_family[request_id]],
            )
            by_version[request["version"]].requests.append(request_info)
        return sorted(by_version.values(), key=lambda x: x.version)

    @cached_property
    def active(self):
        by_ap_key = {}
        to_skip = self.missing | {r.request_id for a in self.archived_requests for r in a.requests}
        for prod in self._active_productions.values():
            if prod["RequestState"] != "Active":
                continue
            request_id = prod["RequestID"]
            if request_id in to_skip:
                continue
            for key in self.request_id_to_ap_keys[request_id]:
                if key not in by_ap_key:
                    by_ap_key[key] = self._raw_ap_infos[key].model_copy(deep=True)
                request = APRequestDebugInfo(
                    request_id=request_id,
                    transforms=[self.transforms.by_id[x] for x in self.transforms.by_family[request_id]],
                )
                by_ap_key[key].requests.append(request)
        return sorted(by_ap_key.values(), key=lambda x: x.created)



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import annotations

__all__ = [
    "to_markdown",
]

from operator import itemgetter

from .prod_analyzer import GroupedProblems, AggregatedLFNDebugInfo
from .base_reasons import LogFileMatch


def _nop_grouper(
    instances: list[AggregatedLFNDebugInfo],
) -> list[tuple[str | None, str | None, list[AggregatedLFNDebugInfo]]]:
    return [(None, None, instances)]


def human_readable_size(size: int) -> str:
    for unit in ["B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB"]:
        if size < 1024:
            return f"{size:.1f} {unit}"
        size /= 1024
    return f"{size:.1f} YiB"


def to_markdown(
    grouped_problems: GroupedProblems, *, is_summary: bool, instance_grouper=_nop_grouper, title: str | None = None
):
    n_lfns = len({(z.transform_id, z.lfn) for x in grouped_problems.values() for y in x.values() for z in y})

    if title is None:
        title = f"Summary of {n_lfns} files"
    body = [f"# {title}\n"]
    for reason_type, matches in sorted(grouped_problems.items()):
        if reason_type:
            body.append(f"## {reason_type}\n")
        for n, (match, instances) in enumerate(sorted(matches.items()), start=1):
            n_reason_lfns = len({(x.lfn, x.transform_id) for x in instances})
            if reason_type == "":
                body.append(f"### {match} in {n_reason_lfns} LFNs\n")
            else:
                if len(matches) > 1:
                    body.append(f"### {reason_type} ({n}) in {n_reason_lfns} LFNs\n")
                body.append(f"```\n{match}\n```\n")
                if reason_type == "KnownCorruptedFile":
                    body.append(
                        "This file is known to contain some corrupt events. It "
                        "likely cannot be recovered for this production however "
                        "we typically don't delete these files as other analyses "
                        "might not depend on the bad branches.\n"
                    )
                else:
                    reason_class = LogFileMatch.name_to_reason[reason_type]  # pylint: disable=unsubscriptable-object
                    if reason_class.explanation:
                        body.append(reason_class.explanation + "\n")
                    if hasattr(reason_class, "issue_url"):
                        body.append(f"See {reason_class.issue_url}.\n")

            for key, key_info, grouped_instances in instance_grouper(instances):
                if key is not None:
                    body.append(f"#### {key}\n")
                if key_info is not None:
                    body.append(f"{key_info}\n")
                body.extend(make_problems_table(grouped_instances, is_summary=is_summary))

    return "\n".join(body)


def make_problems_table(
    grouped_instances: list[AggregatedLFNDebugInfo],
    *,
    is_summary: bool,
    transform_to_group: dict[int, str] | None = None,
) -> list[str]:
    table = [
        "| Transform | LFN | Job ID | Site | Peak Mem | Log URL(s) |",
        "| --------- | --- | ------ | ---- | -------- | ---------- |",
    ]
    if transform_to_group:
        table[0] = "| Group " + table[0]
        table[1] = "| ----- " + table[1]
    if is_summary:
        table[0] += " Occurrences |"
        table[1] += " ----------- |"
    prev_lfn = None
    prev_transform_id = None
    for instance in sorted(grouped_instances, key=lambda x: (x.transform_id, x.lfn, x.job_id)):
        transform_id = ""
        lfn = ""
        if prev_transform_id != instance.transform_id:
            prev_transform_id = instance.transform_id
            transform_id = f"`{instance.transform_id}`"
            lfn = f"`{instance.lfn}`"
        elif prev_lfn != instance.lfn:
            lfn = f"`{instance.lfn}`"
        prev_lfn = instance.lfn
        log_urls = sorted(instance.log_urls, reverse=True)[:3] if is_summary else [instance.log_url]
        if is_summary:
            sites = " ".join(
                f"{f'**{k}**' if k == instance.site else k} ({v})"
                for k, v in sorted(instance.sites.items(), key=itemgetter(1), reverse=True)
            )
        else:
            sites = instance.site

        line = [
            transform_id,
            lfn,
            instance.job_id,
            sites,
            human_readable_size(instance.peak_memory) if instance.peak_memory else "",
            " ".join(f"[here]({url})" for url in log_urls if url),
        ]
        if transform_to_group:
            line.insert(0, transform_to_group[instance.transform_id])
        if is_summary:
            line.append(f"{instance.occurrences} / {instance.attempts}")
        table.append("| " + " | ".join(map(str, line)) + " |")

    table.append("")
    return table



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Module holding ProductionRequestClient class."""
from DIRAC.Core.Base.Client import Client, createClient

from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise, convertToReturnValue


@createClient("ProductionManagement/ProductionRequest")
class ProductionRequestClient(Client):
    """This class expose the methods of the Production Request Service."""

    def __init__(self, url=None, **kwargs):
        """c'tor.

        :param str url: can specify a specific URL
        """
        super().__init__(**kwargs)
        self.setServer("ProductionManagement/ProductionRequest")
        if url:
            self.setServer(url)

    @convertToReturnValue
    def getProductionRequest(self, requestIDList, columns=None):
        """Return the production request for the given request ID list.

        :param list requestIDList: list of request IDs
        :param list columns: list of columns to return, if None all columns are returned except 'RawRequest'
        :return: dictionary with request IDs as keys and their corresponding requests as values
        """
        result = returnValueOrRaise(self.executeRPC(requestIDList, columns, call="getProductionRequest"))
        # Convert the keys to int to reverse the conversion done by JEncode
        return {int(k): v for k, v in result.items()}

    @convertToReturnValue
    def getAllProductionProgress(self):
        """Return all the production progress."""
        result = returnValueOrRaise(self.executeRPC(call="getAllProductionProgress"))
        # Convert the keys to int to reverse the conversion done by JEncode
        return {int(k): {int(k2): v2 for k2, v2 in v.items()} for k, v in result.items()}

    @convertToReturnValue
    def getProductionRequestSummary(self, status, requestType):
        """Return all the production progress."""
        result = returnValueOrRaise(self.executeRPC(status, requestType, call="getProductionRequestSummary"))
        # Convert the keys to int to reverse the conversion done by JEncode
        return {int(k): v for k, v in result.items()}



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Table/object defintions used in the AnalysisProductionsDB"""
import re
from sqlalchemy import (
    Boolean,
    ForeignKeyConstraint,
    Integer,
    Column,
    String,
    JSON,
    DateTime,
    func,
    ForeignKey,
    Float,
    UniqueConstraint,
)
from sqlalchemy.orm import declarative_base, relationship, validates
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql import expression
from sqlalchemy.types import TypeEngine
from pydantic import BaseModel, field_validator, ValidationError

Base = declarative_base()
Base.__table_args__ = {
    "mysql_engine": "InnoDB",
    "mysql_charset": "utf8",
}


class utcnow_add_days(expression.FunctionElement):
    """Sqlalchemy function to return a date now() plus 'days_to_add' days.

    Used to set default datetime as NOW() plus some number of days.

    """

    type: TypeEngine = DateTime()
    inherit_cache: bool = True

    def __init__(self, *args, days_to_add, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if not isinstance(days_to_add, int):
            raise TypeError(f"days_to_add must be an int, got {type(days_to_add)}")
        self._days_to_add = days_to_add


@compiles(utcnow_add_days, "mysql")
def mysql_utcnow_add_days(element, compiler, **kw) -> str:
    """Sqlalchemy function for mysql rendering of utcnow_add_days.

    Part of utcnow_add_days.
    """
    return f"DATE_ADD( UTC_TIMESTAMP, INTERVAL {element._days_to_add} DAY)"


@compiles(utcnow_add_days, "sqlite")
def sqlite_utcnow_add_days(element, compiler, **kw) -> str:
    """Sqlalchemy function for sqlite rendering of utcnow_add_days.

    Part of utcnow_add_days.
    """
    return f"DATE(DATETIME('now'), '+{element._days_to_add} days')"


class Request(Base):
    __tablename__ = "ap_requests"

    VALID_STATES = [
        "waiting",
        "active",
        "replicating",
        "ready",
    ]
    request_id = Column(Integer, primary_key=True)
    filetype = Column(String(64), primary_key=True)
    state = Column(String(16), nullable=False, default="waiting")
    progress = Column(Float, nullable=True)
    last_state_update = Column(
        DateTime(timezone=True),
        nullable=False,
        server_default=func.now(),  # pylint: disable=not-callable
        onupdate=func.now(),  # pylint: disable=not-callable
    )

    name = Column(String(256), nullable=False)
    version = Column(String(64), nullable=False)
    auto_tags = relationship("AutoTag", back_populates="request", lazy="selectin")
    # TODO: Use the mutable extension and validate this object better?
    extra_info = Column(JSON(), nullable=False, default=lambda: {"transformations": []})

    @validates("name")
    def convert_lower(self, key, value):
        return value.lower()

    def toDict(self):
        result = {
            "name": self.name,
            "version": self.version,
            "request_id": self.request_id,
            "filetype": self.filetype,
            "state": self.state,
            "last_state_update": self.last_state_update,
            "transformations": self.extra_info["transformations"],
        }
        if self.progress is not None:
            result["progress"] = self.progress
        if "jira_task" in self.extra_info:
            result["jira_task"] = self.extra_info["jira_task"]
        if "merge_request" in self.extra_info:
            result["merge_request"] = self.extra_info["merge_request"]
        return result


class AnalysisSample(Request):
    __tablename__ = "ap_analysis_samples"

    sample_id = Column(Integer, primary_key=True)
    wg = Column(String(16), nullable=False)
    analysis = Column(String(256), nullable=False)
    request_id = Column(Integer, nullable=False)
    filetype = Column(String(64), nullable=False)

    __table_args__ = (
        ForeignKeyConstraint(["request_id", "filetype"], ["ap_requests.request_id", "ap_requests.filetype"]),
    )

    tags = relationship("Tag", back_populates="sample", lazy="selectin")
    publications = relationship("Publication", back_populates="sample", lazy="selectin")

    # Allow this table to be temporally versioned
    validity_start = Column(
        DateTime(timezone=False), nullable=False, server_default=func.now()  # pylint: disable=not-callable
    )
    validity_end = Column(DateTime(timezone=False), nullable=True)

    housekeeping_interaction_due = Column(DateTime(timezone=False), server_default=utcnow_add_days(days_to_add=90))

    @validates("wg", "analysis")
    def convert_lower(self, key, value):
        return value.lower()

    def toDict(self):
        result = Request.toDict(self)
        result.update(
            {
                "wg": self.wg,
                "analysis": self.analysis,
                "sample_id": self.sample_id,
                "owners": [],
                "validity_start": self.validity_start,
                "validity_end": self.validity_end,
                "housekeeping_interaction_due": self.housekeeping_interaction_due,
                "publications": self.publications or [],
            }
        )
        return result


class User(Base):
    __tablename__ = "ap_users"

    username = Column(String(256), nullable=False, primary_key=True)
    active_since = Column(DateTime(timezone=False), nullable=False, default=func.now())  # pylint: disable=not-callable
    active_until = Column(DateTime(timezone=False), nullable=True)

    n_requests = Column(Integer, nullable=False, default=0)
    accepted_agreement = Column(DateTime(timezone=False), nullable=True, default=None)
    trusted = Column(Boolean, nullable=False, default=False)

    # Links the user to Fence as the username is not used there
    person_id = Column(Integer, nullable=True)

    @validates("username")
    def convert_lower(self, key, value):
        if value is None:
            return None
        return value.lower()

    def toDict(self):
        return {
            "username": self.username,
            "active_since": self.active_since,
            "active_until": self.active_until,
            "n_requests": self.n_requests,
            "accepted_agreement": self.accepted_agreement,
            "trusted": self.trusted,
            "person_id": self.person_id,
        }


class AnalysisOwner(Base):
    __tablename__ = "ap_analysis_owners"

    wg = Column(String(16), primary_key=True)
    analysis = Column(String(256), primary_key=True)
    username = Column(String(256), ForeignKey("ap_users.username"), primary_key=True)

    validity_start = Column(
        DateTime(timezone=False),
        nullable=False,
        primary_key=True,
        server_default=func.now(),  # pylint: disable=not-callable
    )
    validity_end = Column(DateTime(timezone=False), nullable=True)
    created_at = Column(
        DateTime(timezone=False), nullable=False, server_default=func.now()  # pylint: disable=not-callable
    )

    # Add relationship to User
    user = relationship("User", foreign_keys=[username])

    @validates("username", "wg", "analysis")
    def convert_lower(self, key, value):
        return value.lower()

    def toDict(self):
        return {
            "wg": self.wg,
            "analysis": self.analysis,
            "username": self.username,
            "validity_start": self.validity_start.isoformat() if self.validity_start else None,
            "validity_end": self.validity_end.isoformat() if self.validity_end else None,
            "created_by": self.created_by,
            "created_at": self.created_at.isoformat() if self.created_at else None,
        }


class Publication(Base):
    __tablename__ = "ap_publications"

    id = Column(Integer, primary_key=True)
    number = Column(String(64), nullable=False)
    sample_id = Column(Integer, ForeignKey("ap_analysis_samples.sample_id"), nullable=False)
    sample = relationship("AnalysisSample", back_populates="publications", lazy="selectin")

    __table_args__ = (UniqueConstraint("number", "sample_id", name="_pubnumber_sample_id_no_duplicates"),)

    @validates("number")
    def convert_upper(self, key, value):
        return value.upper()


class Tag(Base):
    __tablename__ = "ap_tags"

    id = Column(Integer, primary_key=True)
    name = Column(String(64), nullable=False)
    value = Column(String(64), nullable=False)
    sample_id = Column(Integer, ForeignKey("ap_analysis_samples.sample_id"), nullable=False)
    sample = relationship("AnalysisSample", back_populates="tags", lazy="joined")

    # Allow this table to be temporally versioned
    validity_start = Column(
        DateTime(timezone=False), nullable=False, server_default=func.now()  # pylint: disable=not-callable
    )
    validity_end = Column(DateTime(timezone=False), nullable=True)

    @validates("name", "value")
    def convert_lower(self, key, value):
        return value.lower()


class AutoTag(Base):
    __tablename__ = "ap_auto_tags"

    id = Column(Integer, primary_key=True)
    name = Column(String(64), nullable=False)
    value = Column(String(64), nullable=False)
    # composite foreign keys
    request_id = Column(Integer, nullable=False)
    filetype = Column(String(64), nullable=False)
    request = relationship("Request", back_populates="auto_tags", lazy="joined", enable_typechecks=False)

    __table_args__ = (
        ForeignKeyConstraint(["request_id", "filetype"], ["ap_requests.request_id", "ap_requests.filetype"]),
    )

    @validates("name", "value")
    def convert_lower(self, key, value):
        return value.lower()


class ActionsInputModel(BaseModel):
    """
    Model for dataset actions input.
    """

    @staticmethod
    def validate_publication_number(pub_no: str) -> str:
        """
        Validate publication number format.

        Accepts:
        - LHCb publication format: LHCb-{TYPE}-YYYY-XXX where TYPE is PAPER, CONF, DP, THESIS, FIGURE, or ANA
        - DOI URL format: https://doi.org/... or http://dx.doi.org/...

        Args:
            pub_no: Publication number string to validate

        Returns:
            The validated publication number (stripped of whitespace)

        Raises:
            ValueError: If the publication number format is invalid
        """
        if not pub_no or not pub_no.strip():
            raise ValueError("Publication number cannot be empty")

        trimmed = pub_no.strip()

        # Check length
        if len(trimmed) > 64:
            raise ValueError(f"Publication number too long (>64 chars): {trimmed}")

        # DOI URL format: https://doi.org/... or http://dx.doi.org/...
        doi_url_pattern = re.compile(r"^https?://(dx\.)?doi\.org/.+$", re.IGNORECASE)
        if doi_url_pattern.match(trimmed):
            return trimmed

        # LHCb publication formats: LHCb-PAPER-YYYY-XXX, LHCb-CONF-YYYY-XXX, LHCb-DP-YYYY-XXX,
        # LHCb-THESIS-YYYY-XXX, LHCb-FIGURE-YYYY-XXX, LHCb-ANA-YYYY-XXX
        lhcb_pattern = re.compile(r"^LHCb-(PAPER|CONF|DP|THESIS|FIGURE|ANA)-\d{4}-\d{3}$", re.IGNORECASE)
        if lhcb_pattern.match(trimmed):
            return trimmed

        # Provide helpful error messages
        if trimmed.startswith("http"):
            raise ValueError("Only https://doi.org/ or http://dx.doi.org/ URLs are accepted")

        if not trimmed.startswith("LHCb-"):
            raise ValueError("Publication number must start with 'LHCb-' or be a DOI URL")

        if not re.search(r"LHCb-(PAPER|CONF|DP|THESIS|FIGURE|ANA)", trimmed, re.IGNORECASE):
            raise ValueError("Publication type must be PAPER, CONF, DP, THESIS, FIGURE, or ANA")

        if not re.search(r"\d{4}", trimmed):
            raise ValueError("Publication number must include 4-digit year")

        if not re.search(r"\d{3}$", trimmed):
            raise ValueError("Publication number must end with 3-digit number")

        raise ValueError(
            f"Invalid publication number format: {trimmed}. Expected format: LHCb-{{TYPE}}-YYYY-XXX or https://doi.org/..."
        )

    class ReferSample(BaseModel, extra="ignore"):
        sample_id: int
        add_publications: list[str] | None = None

        @field_validator("add_publications")
        @classmethod
        def validate_publications(cls, v):
            """Validate each publication number in the list"""
            if v is None:
                return v
            validated = []
            errors = []
            for pub_no in v:
                try:
                    validated.append(ActionsInputModel.validate_publication_number(pub_no))
                except ValueError as e:
                    errors.append(str(e))

            if errors:
                raise ValueError(f"Invalid publication numbers: {'; '.join(errors)}")

            return validated

    # archive
    delete: list[ReferSample] = []

    # reminder extensions
    one_month: list[ReferSample] = []
    three_months: list[ReferSample] = []
    six_months: list[ReferSample] = []



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
########################################################################
"""DIRAC Basic Oracle Class It provides access to the basic Oracle methods in a
multithread-safe mode using oracledb connection pooling for connection reuse.

These are the coded methods:

__init__( user, passwd, tns, [maxQueueSize=100] )

Initializes the connection pool and tries to connect to the DB server.
"maxQueueSize" defines the maximum number of connections in the pool.


_except( methodName, exception, errorMessage )

Helper method for exceptions: the "methodName" and the "errorMessage"
are printed with ERROR level, then the "exception" is printed (with
full description if it is a Oracle Exception) and S_ERROR is returned
with the errorMessage and the exception.


_connect()

Creates the connection pool and sets the _connected flag to True upon success.
Returns S_OK or S_ERROR.


query( cmd, [conn] )

Executes SQL command "cmd".
Gets a connection from the pool (or uses the provided connection),
automatically returning it to the pool when done.
Returns S_OK with fetchall() out in Value or S_ERROR upon failure.


_getConnection()

Gets a connection from the pool.
Returns S_OK with connection in Value or S_ERROR
the calling method is responsible for closing this connection once it is no
longer needed.
"""
from contextlib import contextmanager

import oracledb
from oracledb import STRING as oracledb_STRING  # pylint: disable=no-name-in-module
from oracledb import NUMBER as oracledb_NUMBER  # pylint: disable=no-name-in-module

from DIRAC import gLogger
from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities.ReturnValues import DReturnType


gInstancesCount = 0
gModeFixed = False

maxConnectRetry = 100
maxArraysize = 5000  # max allowed


class OracleDB:
    """Basic multithreaded DIRAC Oracle Client Class."""

    def __init__(
        self, userName, password="", tnsEntry="", confDir="", mode="", maxQueueSize=100, call_timeout_ms=None, **kwargs
    ):
        """set Oracle connection parameters and try to connect."""
        global gInstancesCount
        gInstancesCount += 1

        self.__initialized = False
        self._connected = False
        self.call_timeout_ms = call_timeout_ms

        if "logger" not in dir(self):
            self.logger = gLogger.getSubLogger("Oracle")

        # let the derived class decide what to do with if is not 1
        self._threadsafe = oracledb.threadsafety
        self.logger.debug(f"thread_safe = {self._threadsafe}")

        self.__checkQueueSize(maxQueueSize)

        self.__connect_kwargs = kwargs | dict(
            user=userName,
            password=password,
            dsn=tnsEntry,
            config_dir=confDir,
        )
        self.__mode = mode
        self.__maxQueueSize = maxQueueSize
        self.__pool = None

        self.__initialized = True
        self._connect()

        if not self._connected:
            raise RuntimeError("Can not connect, exiting...")

        self.logger.info("===================== Oracle =====================")
        self.logger.info("User:           " + self.__connect_kwargs.get("user"))
        self.logger.info("TNS:            " + self.__connect_kwargs.get("dsn"))
        self.logger.info("Pool size:      " + str(maxQueueSize))
        self.logger.info("==================================================")

    def __del__(self):
        global gInstancesCount

        if self.__initialized and self.__pool:
            try:
                self.__pool.close()
                self.logger.debug("Connection pool closed")
            except Exception as e:
                self.logger.debug(f"Error closing connection pool: {e}")

    @staticmethod
    def __checkQueueSize(maxQueueSize):
        """the size of the connection pool is limited."""

        if maxQueueSize <= 0:
            raise Exception("OracleDB.__init__: maxQueueSize must positive")
        try:
            test = maxQueueSize - 1
        except TypeError:
            raise TypeError(f"OracleDB.__init__: wrong type for maxQueueSize {type(maxQueueSize)}")

    def _except(self, methodName, x, err):
        """print Oracle error or exeption return S_ERROR with Exception."""

        try:
            raise x
        except oracledb.Error as e:
            self.logger.error(f"{methodName}: {err}", str(e))
            return S_ERROR(f"{err}: ( {e} )")
        except Exception as x:
            self.logger.error(f"{methodName}: {err}", str(x))
            return S_ERROR(f"{err}: ({x})")

    def _connect(self):
        """create connection pool and set connected flag to True upon success."""
        self.logger.debug("_connect:", self._connected)
        if self._connected:
            return S_OK()

        self.logger.debug(
            "_connect: Attempting to create connection pool", f"for user {self.__connect_kwargs.get('user')}."
        )
        try:
            self.__createPool()
            self.logger.debug("_connect: Connection pool created.")
            self._connected = True
            return S_OK()
        except Exception as x:
            return self._except("_connect", x, "Could not create connection pool.")

    def query(self, cmd, conn=False, params=[], kwparams={}, pre_inserts=[]):
        """execute Oracle query command return S_OK structure with fetchall result
        as tuple it returns an empty tuple if no matching rows are found return
        S_ERROR upon error.

        Use of params and kwparams to pass bind variabes is strongly encouraged
        to prevent SQL injection and improve performance. See the python-oracledb
        documentation for more information.

        :param str cmd: the SQL string to be executed
        :param conn: the connection to use, optional
        :param list params: positional bind variables to pass to oracledb.Cursor.execute
        :param dict kwparams: named bind variables to pass to oracledb.Cursor.execute
        :param list pre_inserts: list of tuples with the query and data to be inserted before the main query
        :param call_timeout_ms: optional timeout in milliseconds for the query execution
        """
        self.logger.debug("query:", f"{cmd!r} {params!r} {kwparams!r}")

        with self.__getConnection(conn) as retDict:
            if not retDict["OK"]:  # pylint: disable=unsubscriptable-object
                return retDict
            connection = retDict["Value"]  # pylint: disable=unsubscriptable-object

            try:
                cursor = connection.cursor()
                cursor.arraysize = maxArraysize

                for pre_insert_query, pre_insert_data in pre_inserts:
                    self.logger.debug(
                        "query: Pre-inserting data into temp table", f"{pre_insert_query!r} {pre_insert_data!r}"
                    )
                    cursor.executemany(pre_insert_query, pre_insert_data)

                if cursor.execute(cmd, *params, **kwparams):
                    res = cursor.fetchall()
                else:
                    res = ()

                # Log the result limiting it to just 10 records
                if len(res) < 10:
                    self.logger.debug("query: Records returned", res)
                else:
                    self.logger.debug(
                        "query: First 10 records returned out of",
                        f"{len(res)}: {res[:10]} ...",
                    )

                retDict = S_OK(res)
            except Exception as x:
                self.logger.debug("query:", cmd)
                retDict = self._except("query", x, "Execution failed.")
                self.logger.debug("Start Rollback transaction")
                connection.rollback()
                self.logger.debug("End Rollback transaction")

            try:
                connection.commit()
                cursor.close()
            except Exception:
                pass

        return retDict

    def executeStoredProcedure(self, packageName, parameters, output=True, array=None, conn=False):
        """executes a stored procedure."""
        self.logger.debug("_query:", packageName + "(" + str(parameters) + ")")

        with self.__getConnection(conn) as retDict:
            if not retDict["OK"]:  # pylint: disable=unsubscriptable-object
                return retDict
            connection = retDict["Value"]  # pylint: disable=unsubscriptable-object

            try:
                cursor = connection.cursor()
                result = None
                results = None
                if array:
                    fArray = array[0]
                    if isinstance(fArray, str):
                        result = cursor.arrayvar(oracledb_STRING, array)
                        parameters += [result]
                    elif isinstance(fArray, int):
                        result = cursor.arrayvar(oracledb_NUMBER, array)
                        parameters += [result]
                    elif isinstance(fArray, list):
                        for i in array:
                            if isinstance(i, (bool, str, int)):
                                parameters += [i]
                            elif i:
                                if isinstance(i[0], str):
                                    result = cursor.arrayvar(oracledb_STRING, i)
                                    parameters += [result]
                                elif isinstance(i[0], int):
                                    result = cursor.arrayvar(oracledb_NUMBER, i)
                                    parameters += [result]
                                else:
                                    return S_ERROR("The array type is not supported!!!")
                            else:
                                result = cursor.arrayvar(oracledb_STRING, [], 0)
                                parameters += [result]
                    else:
                        return S_ERROR("The array type is not supported!!!")
                if output:
                    result = connection.cursor()
                    result.arraysize = maxArraysize  # 500x faster!!
                    parameters += [result]
                    cursor.callproc(packageName, parameters)
                    results = result.fetchall()
                else:
                    cursor.callproc(packageName, parameters)
                retDict = S_OK(results)
            except Exception as x:
                self.logger.debug("query:", packageName + "(" + str(parameters) + ")")
                retDict = self._except("query", x, "Execution failed.")
                connection.rollback()

            try:
                cursor.close()
            except Exception as ex:
                self._except("executeStoredProcedure:", ex, "Failed to close cursor")

        return retDict

    def executeStoredFunctions(self, packageName, returnType, parameters=None, conn=False):
        """executes a stored function."""
        if parameters is None:
            parameters = []
        with self.__getConnection(conn) as retDict:
            if not retDict["OK"]:  # pylint: disable=unsubscriptable-object
                return retDict
            connection = retDict["Value"]  # pylint: disable=unsubscriptable-object
            try:
                cursor = connection.cursor()
                cursor.arraysize = maxArraysize
                result = cursor.callfunc(packageName, returnType, parameters)
                retDict = S_OK(result)
            except Exception as x:
                self.logger.debug(f"_query: {packageName} ({parameters})")
                retDict = self._except("_query", x, "Execution failed.")
                connection.rollback()

            try:
                cursor.close()
            except Exception as ex:
                self._except("executeStoredFunctions:", ex, "Failed to close cursor")
        return retDict

    def __createPool(self):
        """Create a connection pool."""
        global gModeFixed
        self.logger.debug("__createPool:")
        if not gModeFixed:
            gModeFixed = True
            if self.__mode == "":
                try:
                    # Test connection in thin mode first
                    test_conn = oracledb.connect(**self.__connect_kwargs)
                    test_conn.close()
                except Exception as ex:
                    self.logger.exception()
                    self.logger.debug(f"Thin mode has failed: {ex}, we will try thick")
                    self.__mode = "Thick"
                else:
                    self.logger.debug("Using implicit thin mode")
            if self.__mode == "Thick":
                oracledb.init_oracle_client(config_dir=self.__connect_kwargs.get("config_dir"))
            self.logger.debug(f'Using {"thin" if oracledb.is_thin_mode() else "thick"} mode')

        self.__pool = oracledb.create_pool(
            min=1, max=self.__maxQueueSize, increment=1, timeout=120, **self.__connect_kwargs
        )
        self.logger.debug(f"Created connection pool with max {self.__maxQueueSize} connections")

    @contextmanager
    def __getConnection(self, conn=None) -> DReturnType:
        """Context manager to get a connection from the pool.

        If a connection is provided, it will be used instead of acquiring a new one."""
        self.logger.debug("__getConnection:")
        if conn:
            yield conn
            return

        if not self.__pool:
            yield S_ERROR("Connection pool not initialized")
            return

        try:
            with self.__pool.acquire() as connection:
                self.logger.debug("__getConnection: Got a connection from pool")
                if self.call_timeout_ms is not None:
                    orig_call_timeout = connection.call_timeout
                    connection.call_timeout = self.call_timeout_ms
                try:
                    yield S_OK(connection)
                finally:
                    if self.call_timeout_ms is not None:
                        connection.call_timeout = orig_call_timeout
        except Exception as x:
            self._except("__getConnection:", x, "Failed to get connection from pool")



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import json
import sys
import re
from datetime import UTC
from typing import Any

from DIRAC import gLogger
from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise, convertToReturnValue
from DIRAC.Core.Base.Script import Script

from LHCbDIRAC.ProductionManagementSystem.Utilities.Models import DataProduction, SimulationProduction, ProductionBase

REQUEST_COLUMNS = [
    "RequestID",
    "ParentID",
    "MasterID",
    "RequestAuthor",
    "RequestName",
    "RequestType",
    "RequestState",
    "RequestPriority",
    "RequestPDG",
    "RequestWG",
    "SimCondition",
    "SimCondID",
    "SimCondDetail",
    "ProPath",
    "ProID",
    "ProDetail",
    "EventType",
    "NumberOfEvents",
    "Description",
    "Comments",
    "Inform",
    "RealNumberOfEvents",
    "IsModel",
    "Extra",
    "RawRequest",
    "RetentionRate",
    "FastSimulationType",
    "HasSubrequest",
]

TRANSFORM_COLUMNS = [
    "TransformationID",
    "TransformationFamily",
    "TransformationName",
    "Status",
]

IGNORE_TRANSFORM_STATES = {"Cleaning", "Cleaned"}


def parseArgs():
    doForReal = False
    validation = False

    @convertToReturnValue
    def enableDoForReal(_):
        nonlocal doForReal
        doForReal = True

    @convertToReturnValue
    def enableValidation(_):
        nonlocal validation
        validation = True

    switches = [
        ("", "do-for-real", "Actually create steps and submit productions", enableDoForReal),
        ("", "validation", "Run validation production", enableValidation),
    ]
    Script.registerSwitches(switches)
    Script.registerArgument("yaml_path: Path to the YAML file containing productions to submit")
    Script.parseCommandLine(ignoreErrors=False)
    requestIDs = [int(x) for x in Script.getPositionalArgs()]
    return requestIDs, doForReal, validation


@Script()
def main():
    from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequestClient import ProductionRequestClient
    from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
    from LHCbDIRAC.ProductionManagementSystem.Utilities.Models import parse_obj as parse_request_dict

    requestIDs, doForReal, validation = parseArgs()

    if not requestIDs:
        raise ValueError("No request IDs provided. Please provide at least one request ID.")

    requests = returnValueOrRaise(ProductionRequestClient().getProductionRequest(requestIDs, columns=REQUEST_COLUMNS))
    if missing := set(requestIDs).symmetric_difference(requests.keys()):
        gLogger.error(f"Some of the provided request IDs were not found in the system. Missing IDs: {missing}")
        sys.exit(1)

    transforms = returnValueOrRaise(
        TransformationClient().getTransformations(
            {"TransformationFamily": list(requestIDs)}, columns=TRANSFORM_COLUMNS
        ),
    )
    # By default, the TransformationFamily is returned as a string, but we need it as an int to match the request IDs.
    for transform in transforms:
        transform["TransformationFamily"] = int(transform["TransformationFamily"])

    transforms_by_family = {}
    family_to_generations = {}
    for transform in transforms:
        if match := re.search(r"_(\d+)\.xml$", transform["TransformationName"]):
            family_to_generations.setdefault(transform["TransformationFamily"], set()).add(int(match.group(1)))
        else:
            raise NotImplementedError(
                f"Transformation {transform['TransformationName']} does not have a valid name suffix. "
                "Expected the transformation name to end with _<int>.xml"
            )
        if transform["Status"] in IGNORE_TRANSFORM_STATES:
            continue
        transforms_by_family.setdefault(transform["TransformationFamily"], []).append(transform)

    if transforms_by_family:
        msg = f"Found transformations for the provided request IDs: {transforms_by_family!r}."
        msg += " If you want to launch the requests, please clean them up first."
        if doForReal:
            raise NotImplementedError(msg)
        gLogger.error(msg)

    for request_id, data in requests.items():
        append_name = str(max(family_to_generations.get(request_id, {0})) + 1)
        request = parse_request_dict(json.loads(data["RawRequest"]))

        subrequests = []
        result = returnValueOrRaise(ProductionRequestClient().getProductionRequestList(request_id, "", "ASC", 0, 0, {}))
        for row in result["Rows"]:
            subrequests.append((row["RequestID"], row["ParentID"] or None, row["EventType"]))
        if not subrequests:
            subrequests.append((request_id, None, None))

        set_state(request_id, "Accepted", dryRun=not doForReal)
        for request_id, parent_id, event_type in subrequests:
            gLogger.info(f"Launching production for request ({parent_id}) with append name {append_name}.")
            _launch_production(
                request_id,
                parent_id,
                request,
                data,
                append_name,
                event_type,
                dryRun=not doForReal,
                validation=validation,
            )
        # Simulation productions stay in Accepted until testing is done.
        if not isinstance(request, SimulationProduction):
            set_state(request_id, "Active", dryRun=not doForReal)


def set_state(request_id: int, state: str, dryRun: bool = False) -> bool:
    """Set the state of a production request."""
    from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequestClient import ProductionRequestClient

    if dryRun:
        gLogger.info(f"Dry run mode enabled, not changing state of request {request_id} to {state}.")
        return True

    gLogger.info(f"Changing state of request {request_id} to {state}.")
    return returnValueOrRaise(ProductionRequestClient().updateProductionRequest(request_id, {"RequestState": state}))


def _launch_production(
    request_id: int,
    parent_id: int | None,
    request: ProductionBase,
    data: dict[str, Any],
    append_name: str,
    event_type: None | str,
    validation: bool,
    dryRun: bool,
) -> list[int]:
    from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequest import (
        ProductionRequest,
    )
    from LHCbDIRAC.ProductionManagementSystem.Utilities.ModelCompatibility import (
        configure_input,
    )
    from LHCbDIRAC.TransformationSystem.Client.TransformationClient import (
        TransformationClient,
    )

    pr = ProductionRequest()
    pr.requestID = str(request_id)
    pr.appendName = append_name
    pr.visibility = "Yes"
    pr.disableOptionalCorrections = True
    pr.bkSampleMax = request.input_dataset.launch_parameters.sample_max_md5
    pr.bkSampleSeed = request.input_dataset.launch_parameters.sample_seed_md5
    if isinstance(request, DataProduction):
        assert event_type is None, event_type
        assert parent_id is None, parent_id
        configure_input(
            pr,
            data,
            runs=request.input_dataset.launch_parameters.run_numbers,
            startRun=request.input_dataset.launch_parameters.start_run,
            endRun=request.input_dataset.launch_parameters.end_run,
        )
        pr.prodGroup = f"{pr.processingPass}/{json.loads(data['ProDetail'])['pDsc']}"[:50]
        pr.extraModulesList = ["FileUsage"]
    elif isinstance(request, SimulationProduction):
        pr.eventType = request.event_types
        pr.requestID = request_id
        pr.parentRequestID = parent_id
        pr.prodGroup = json.loads(data["ProDetail"])["pDsc"]
    else:
        raise NotImplementedError(request)
    pr.outConfigName = pr.configName

    if validation:
        if not isinstance(request, SimulationProduction):
            if request.input_dataset.launch_parameters.start_run is not None:
                raise ValueError("Expected start_run to be None for validation requests")
            if request.input_dataset.launch_parameters.end_run is not None:
                raise ValueError("Expected end_run to be None for validation requests")
            if not request.input_dataset.launch_parameters.run_numbers:
                raise ValueError("Expected run_numbers to be provided for validation requests")
        pr.outConfigName = "validation"

    configure_steps(pr, data, request)
    ourStepsInProds = pr.stepsInProds

    if dryRun:
        gLogger.info(
            "Dry run mode enabled, to actually launch the production, pass --do-for-real",
        )
        return []

    transform_ids = returnValueOrRaise(pr.buildAndLaunchRequest())
    name = f"{request_id} ({parent_id})" if parent_id else str(request_id)
    gLogger.info(f"Production for request {name} launched successfully.")

    if ourStepsInProds != pr.stepsInProds:
        raise ValueError(
            "Something went horribly wrong: stepsInProds was changed after buildAndLaunchRequest. "
            "Check that applyOptionalCorrections didn't meddle in our business!\n"
            f"The following transformations should be stopped and cleaned: {transform_ids!r}"
        )

    if isinstance(request, DataProduction):
        # If keep_running is False, we need to update the end date of the transformation
        # so files are only added if they exist at the time of launching the request
        new_query = None
        tid = transform_ids[0]
        if end_date := request.input_dataset.launch_parameters.end_date:
            # If the input dataset has an end date, we use that as the end date for the transformation
            new_query = {"EndDate": end_date.astimezone(UTC).strftime("%Y-%m-%d %H:%M:%S")}

        if request.input_dataset.launch_parameters.sample_max_md5:
            new_query = (new_query or {}) | {
                "SampleMax": request.input_dataset.launch_parameters.sample_max_md5,
                "SampleSeedMD5": request.input_dataset.launch_parameters.sample_seed_md5,
            }

        if new_query is not None:
            tc = TransformationClient()
            new_query = returnValueOrRaise(tc.getBookkeepingQuery(tid)) | new_query
            returnValueOrRaise(tc.deleteBookkeepingQuery(tid))
            returnValueOrRaise(tc.addBookkeepingQuery(tid, new_query))
            print(f"Modified input query for transformation {tid} to: {new_query}")

    return transform_ids


def find_previousProds(trf_idx, request: ProductionBase, stepsInProds):
    trf = request.submission_info.processing_transforms[trf_idx]
    if trf_idx == 0:
        return None
    from_step_idxs = {x.step_idx for x in request.steps[trf.steps[0]].input}
    if len(from_step_idxs) != 1:
        raise NotImplementedError("Please don't")
    from_step_idx = from_step_idxs.pop()
    for trf_idx_, stepsInProd in enumerate(stepsInProds):
        if from_step_idx + 1 in stepsInProd:
            return trf_idx_ + 1
    raise NotImplementedError("Something went very wrong")


def configure_steps(pr, data, request: ProductionBase):
    pd = json.loads(data["ProDetail"])
    steps_stepIDs = {int(k[1 : -len("Step")]): int(v) for k, v in pd.items() if k.endswith("Step")}
    pr.stepsList = [int(v) for k, v in sorted(steps_stepIDs.items())]

    # The EventTimeout isn't actually stored in the bookkeeping so take it from the raw request
    for dirac_step in request.steps:
        assert dirac_step.id is not None, "All steps must have an ID assigned before launching the production"
        if dirac_step.event_timeout is not None:
            pr.eventTimeouts[dirac_step.id] = dirac_step.event_timeout

    pr.previousProds = []
    for trf_idx, trf in enumerate(request.submission_info.processing_transforms):
        pr.prodsTypeList += [trf.type]
        # which stepIDs are in this transform
        pr.stepsInProds += [[x + 1 for x in trf.steps]]  # this is actually 1-indexed in DIRAC 😱😭
        # where do the outputs get uploaded to
        pr.outputSEs += [trf.output_se]
        # specifies how we access our input files in the transform
        pr.inputDataPolicies += [trf.input_data_policy]
        # the input file plugin to use for the transform
        pr.plugins += [trf.input_plugin]
        # Just the number of steps in this transform.
        pr.outputFileSteps += [len(trf.steps)]
        # the input file group size to the transform.
        pr.groupSizes += [trf.group_size]
        # whether the output is going to be saved?
        output_visibilities = {}
        for trf_step_idx in trf.steps:
            visibilities = {x.visible for x in request.steps[trf_step_idx].output}
            if len(visibilities) != 1:
                raise NotImplementedError("No idea how to make this work...")
            output_visibilities[str(trf_step_idx + 1)] = {True: "Y", False: "N"}[visibilities.pop()]
        pr.outputVisFlag.append(output_visibilities)

        # previousProds is a list of indices (ONE-INDEXED!!!)
        # each pointing to the transformation to take input from
        # referenced in the stepsInProds list.
        # we need to do this otherwise the transforms will not find the correct
        # output to use as input.
        # e.g.
        # 1 = take input from (1-1=0)th transform referenced in stepsInProds
        # 2 = take input from (2-1=1)st transform referenced in stepsInProds
        # 3 = take input from (3-1=2)st transform referenced in stepsInProds
        # n = take input from (n-1)th transform referenced in stepsInProds
        pr.previousProds.append(find_previousProds(trf_idx, request, pr.stepsInProds))
        if trf_idx > 0:
            pr.bkQueries.append("fromPreviousProd")
        elif not pr.bkQueries:
            # Only set the first bkQuery if not already set by configure_input
            # (DataProduction sets it to "Full" or "fromPreviousProd", SimulationProduction leaves it empty)
            pr.bkQueries.append("")

        # left blank
        pr.targets += [""]
        # left blank
        pr.specialOutputSEs += [{}]
        # whether the input files should be removed when used by
        # the transform - need to be very careful with this
        pr.removeInputsFlags += [trf.remove_inputs_flags]
        # leave as is
        pr.priorities += [trf.priority]
        # left blank
        pr.inputs += [[]]
        # ???
        pr.outputFileMasks += [trf.output_file_mask]
        # left as is
        pr.multicore += [str(trf.multicore)]
        # ???
        pr.outputModes += [trf.output_mode]
        # we process all events
        pr.events += [trf.events]
        # leave as is as not relevant
        pr.ancestorDepths += [trf.ancestor_depth]
        # extra input sandboxes to be used in the transform
        pr.inputSandboxes += [trf.extra_sandboxes or []]
        # compression is not handled at this point but in our
        # application config
        pr.compressionLvl += ["LOW"]
        # left as is
        pr.cpus += [trf.cpu]
        # processors
        pr.processors += [trf.processors]

    # DO NOT REMOVE THIS!
    if pr.removeInputsFlags[0] is not False:
        raise NotImplementedError("The first transformation should never remove inputs...")
    if len(pr.previousProds) != len(pr.prodsTypeList):
        raise ValueError("The number of previous productions does not match the number of production types.")


if __name__ == "__main__":
    main()
